import abc
import io
import itertools
import os
import unittest
import random
import tempfile
import threading

import pyhmmer
from pyhmmer.easel import Alphabet, SequenceFile, DigitalSequence, DigitalSequenceBlock, TextSequence, TextSequenceBlock, MSAFile, DigitalMSA
from pyhmmer.plan7 import Background, Builder, Pipeline, HMMFile, TopHits, OptimizedProfileBlock, Profile, LongTargetsPipeline
from pyhmmer.errors import AlphabetMismatch

from .. import __name__ as __package__
from .utils import HMMER_FOLDER, resource_files


@unittest.skipUnless(resource_files, "importlib.resources.files not available")
class TestSearchPipeline(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.alphabet = Alphabet.amino()

        cls.reference_path = resource_files(__package__).joinpath("data", "seqs", "938293.PRJEB85.HG003687.faa")
        with SequenceFile(cls.reference_path, digital=True, alphabet=cls.alphabet) as f:
            cls.references = f.read_block()

        cls.msa_path = resource_files(__package__).joinpath("data", "msa", "LuxC.sto")
        with MSAFile(cls.msa_path, digital=True, alphabet=cls.alphabet) as msa_f:
            cls.msa = msa_f.read()

    def test_search_hmm_alphabet_mismatch(self):
        rng = pyhmmer.easel.Randomness(0)
        pipeline = Pipeline(alphabet=Alphabet.dna())

        # mismatch between pipeline alphabet and database alphabet
        hmm1 = pyhmmer.plan7.HMM.sample(pipeline.alphabet, 100, rng)
        self.assertRaises(AlphabetMismatch, pipeline.search_hmm, hmm1, self.references)

        # mismatch between pipeline alphabet and query alphabet
        hmm2 = pyhmmer.plan7.HMM.sample(self.alphabet, 100, rng)
        self.assertRaises(AlphabetMismatch, pipeline.search_hmm, hmm2, self.references)

    def test_search_seq_alphabet_mismatch(self):
        pipeline = Pipeline(alphabet=Alphabet.dna())

        # mismatch between pipeline alphabet and database alphabet
        dsq1 = TextSequence(sequence="ATGC").digitize(pipeline.alphabet)
        self.assertRaises(AlphabetMismatch, pipeline.search_seq, dsq1, self.references)

        # mismatch between pipeline alphabet and query alphabet
        dsq2 = TextSequence(sequence="IRGIY").digitize(self.alphabet)
        self.assertRaises(AlphabetMismatch, pipeline.search_seq, dsq2, self.references)

    def test_search_msa_alphabet_mismatch(self):
        pipeline = Pipeline(alphabet=Alphabet.dna())

        # mismatch between pipeline alphabet and query alphabet
        self.assertRaises(AlphabetMismatch, pipeline.search_msa, self.msa, self.references)

        # mismatch between pipeline alphabet and database alphabet
        dsq = TextSequence(sequence="ATGC").digitize(pipeline.alphabet)
        msa = DigitalMSA(pipeline.alphabet, sequences=[dsq], name=b"test")
        self.assertRaises(AlphabetMismatch, pipeline.search_msa, msa, self.references)

    def test_search_hmm_block(self):
        seq = TextSequence(sequence="IRGIYNIIKSVAEDIEIGIIPPSKDHVTISSFKSPRIADT", name=b"seq1")
        bg = Background(self.alphabet)
        hmm, _, _ = Builder(self.alphabet).build(seq.digitize(self.alphabet), bg)
        pipeline = Pipeline(alphabet=self.alphabet)
        hits = pipeline.search_hmm(hmm, self.references)
        self.assertEqual(len(hits), 1)
        self.assertEqual(hits.query.name, hmm.name)
        self.assertEqual(hits.query.accession, hmm.accession)

    def test_search_hmm_file(self):
        seq = TextSequence(sequence="IRGIYNIIKSVAEDIEIGIIPPSKDHVTISSFKSPRIADT", name=b"seq1")
        bg = Background(self.alphabet)
        hmm, _, _ = Builder(self.alphabet).build(seq.digitize(self.alphabet), bg)
        pipeline = Pipeline(alphabet=self.alphabet)
        with SequenceFile(self.reference_path, digital=True, alphabet=self.alphabet) as seqs_file:
            hits = pipeline.search_hmm(hmm, seqs_file)
        self.assertEqual(len(hits), 1)
        self.assertEqual(hits.query.name, hmm.name)
        self.assertEqual(hits.query.accession, hmm.accession)

    def test_search_hmm_unnamed(self):
        # make sure `Pipeline.search_hmm` doesn't crash when given an HMM with no name
        rng = pyhmmer.easel.Randomness()
        hmm = pyhmmer.plan7.HMM.sample(self.alphabet, 100, rng)
        hmm.name = b"test"
        hmm.accession = None
        pipeline = Pipeline(alphabet=self.alphabet)
        hits = pipeline.search_hmm(hmm, self.references)
        self.assertEqual(hits.query.name, b"test")
        self.assertIs(hits.query.accession, None)

    def test_search_hmm_unsupported(self):
        pipeline = Pipeline(alphabet=self.alphabet)
        with self.assertRaises(TypeError):
            with SequenceFile(self.reference_path, digital=True, alphabet=self.alphabet) as seqs_file:
                hits = pipeline.search_hmm(object(), seqs_file)

    def test_search_seq_block(self):
        seq = TextSequence(sequence="IRGIYNIIKSVAEDIEIGIIPPSKDHVTISSFKSPRIADT", name=b"seq1", accession=b"SQ001")
        pipeline = Pipeline(alphabet=self.alphabet)
        hits = pipeline.search_seq(seq.digitize(self.alphabet), self.references)
        self.assertEqual(len(hits), 1)
        self.assertEqual(hits.query.name, seq.name)
        self.assertEqual(hits.query.accession, seq.accession) # NOTE: p7_SingleBuilder doesn't copy the accession...

    def test_search_seq_file(self):
        seq = TextSequence(sequence="IRGIYNIIKSVAEDIEIGIIPPSKDHVTISSFKSPRIADT", name=b"seq1", accession=b"SQ001")
        pipeline = Pipeline(alphabet=self.alphabet)
        with SequenceFile(self.reference_path, digital=True, alphabet=self.alphabet) as seqs_file:
            hits = pipeline.search_seq(seq.digitize(self.alphabet), seqs_file)
        self.assertEqual(len(hits), 1)
        self.assertEqual(hits.query.name, seq.name)
        self.assertEqual(hits.query.accession, seq.accession) # NOTE: p7_SingleBuilder doesn't copy the accession...

    def test_search_seq_unsupported(self):
        pipeline = Pipeline(alphabet=self.alphabet)
        with self.assertRaises(TypeError):
            with SequenceFile(self.reference_path, digital=True, alphabet=self.alphabet) as seqs_file:
                hits = pipeline.search_seq(object(), seqs_file)

    def test_search_msa_block(self):
        pipeline = Pipeline(alphabet=self.alphabet)
        hits = pipeline.search_msa(self.msa, self.references)
        self.assertEqual(len(hits), 1)
        self.assertEqual(hits.query.name, self.msa.name)
        self.assertEqual(hits.query.accession, self.msa.accession)

    def test_search_msa_file(self):
        pipeline = Pipeline(alphabet=self.alphabet)
        with SequenceFile(self.reference_path, digital=True, alphabet=self.alphabet) as seqs_file:
            hits = pipeline.search_msa(self.msa, seqs_file)
        self.assertEqual(len(hits), 1)
        self.assertEqual(hits.query.name, self.msa.name)
        self.assertEqual(hits.query.accession, self.msa.accession)

    def test_search_msa_unsupported(self):
        pipeline = Pipeline(alphabet=self.alphabet)
        with self.assertRaises(TypeError):
            with SequenceFile(self.reference_path, digital=True, alphabet=self.alphabet) as seqs_file:
                hits = pipeline.search_msa(object(), seqs_file)

    def test_Z(self):
        seq = TextSequence(sequence="IRGIYNIIKSVAEDIEIGIIPPSKDHVTISSFKSPRIADT", name=b"seq1")
        bg = Background(self.alphabet)
        hmm, _, _ = Builder(self.alphabet).build(seq.digitize(self.alphabet), bg)

        # when Z is None, use the number of target sequences
        pipeline = Pipeline(alphabet=self.alphabet)
        self.assertIs(pipeline.Z, None)
        hits = pipeline.search_hmm(hmm, self.references[:100])
        self.assertEqual(hits.Z, 100)
        self.assertIs(pipeline.Z, None)
        # clearing the pipeline will keep the Z number as None
        pipeline.clear()
        self.assertIs(pipeline.Z, None)

        # when Z is not None, use the given number
        pipeline = Pipeline(alphabet=self.alphabet, Z=25)
        self.assertEqual(pipeline.Z, 25)
        hits = pipeline.search_hmm(hmm, self.references[:100])
        self.assertEqual(pipeline.Z, 25)
        self.assertEqual(hits.Z, 25)
        # clearing the pipeline will keep the Z number
        pipeline.clear()
        self.assertEqual(pipeline.Z, 25)

    def test_bit_cutoffs(self):
        seq = TextSequence(sequence="IRGIYNIIKSVAEDIEIGIIPPSKDHVTISSFKSPRIADT")
        bg = Background(self.alphabet)
        hmm, _, _ = Builder(self.alphabet).build(seq.digitize(self.alphabet), bg)
        pipeline = Pipeline(alphabet=self.alphabet, bit_cutoffs="trusted")

        # without cutoffs, the pipeline is not going to be able to process the hmm
        with self.assertRaises(ValueError):
            hits = pipeline.search_hmm(hmm, self.references)

        # with low cutoffs, the domain is found
        hmm.cutoffs.trusted = (10.0, 10.0)
        hits = pipeline.search_hmm(hmm, self.references)
        self.assertEqual(len(hits), 1)
        self.assertLessEqual(hits[0].score, 75.0)

        # with higher cutoff, no domain is found
        hmm.cutoffs.trusted = (75.0, 75.0)
        hits = pipeline.search_hmm(hmm, self.references)
        self.assertEqual(len(hits), 0)


@unittest.skipUnless(resource_files, "importlib.resources.files not available")
class TestScanPipeline(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.alphabet = Alphabet.amino()

        seq_path = resource_files(__package__).joinpath("data", "seqs", "938293.PRJEB85.HG003687.faa")
        with SequenceFile(seq_path, digital=True, alphabet=cls.alphabet) as f:
            cls.references = f.read_block()

        hmm_file = resource_files(__package__).joinpath("data", "hmms", "txt", "RREFam.hmm")
        with HMMFile(hmm_file) as f:
            cls.hmms = list(f)

    def test_scan_seq_alphabet_mismatch(self):
        pipeline = Pipeline(alphabet=Alphabet.dna())

        # mismatch between pipeline alphabet and query alphabet
        dsq = TextSequence(sequence="IRGIY").digitize(self.alphabet)
        oprofiles = OptimizedProfileBlock(pipeline.alphabet)
        self.assertRaises(AlphabetMismatch, pipeline.scan_seq, dsq, oprofiles)

        # mismatch between pipeline alphabet and database alphabet
        dsq = TextSequence(sequence="ATGC").digitize(pipeline.alphabet)
        oprofiles = OptimizedProfileBlock(self.alphabet)
        self.assertRaises(AlphabetMismatch, pipeline.scan_seq, dsq, oprofiles)

    def test_scan_seq_block(self):
        seq = next(x for x in self.references if x.name == b"938293.PRJEB85.HG003691_78")

        oprofiles = OptimizedProfileBlock(seq.alphabet)
        background = Background(seq.alphabet)
        for hmm in self.hmms:
            profile = Profile(hmm.M, hmm.alphabet)
            profile.configure(hmm, background)
            oprofiles.append(profile.to_optimized())

        pipeline = Pipeline(alphabet=self.alphabet)
        hits = pipeline.scan_seq(seq, oprofiles)
        self.assertEqual(len(hits), 3)  # number found with `hmmscan`

    def test_scan_seq_file(self):
        seq = next(x for x in self.references if x.name == b"938293.PRJEB85.HG003691_78")

        hmm_file = resource_files(__package__).joinpath("data", "hmms", "db", "RREFam.hmm")
        with HMMFile(hmm_file) as f:
            pipeline = Pipeline(alphabet=self.alphabet)
            hits = pipeline.scan_seq(seq, f.optimized_profiles())

        self.assertEqual(len(hits), 3)  # number found with `hmmscan`


@unittest.skipUnless(resource_files, "importlib.resources.files not available")
class TestIteratePipeline(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.alphabet = Alphabet.amino()

        seq_path = resource_files(__package__).joinpath("data", "seqs", "938293.PRJEB85.HG003687.faa")
        with SequenceFile(seq_path, digital=True, alphabet=cls.alphabet) as f:
            cls.references = f.read_block()

    def test_iterate_seq(self):
        # check that `Pipeline.iterate_seq` produces consistent results
        # compared to a 'gold standard' run of `jackhmmer`.

        query_path = resource_files(__package__).joinpath("data", "seqs", "LuxC.faa")
        with SequenceFile(query_path, digital=True, alphabet=self.alphabet) as f:
            query = next(seq for seq in f if b"P12748" in seq.name)

        pipeline = Pipeline(alphabet=self.alphabet, incE=0.001, incdomE=0.001)
        search_iterator = pipeline.iterate_seq(query, self.references)

        for n in range(3):
            self.assertEqual(search_iterator.iteration, n)
            iteration = next(search_iterator)

            self.assertEqual(search_iterator.iteration, n+1)
            self.assertEqual(iteration.iteration, n+1)

            msa_path = resource_files(__package__).joinpath("data", "jackhmmer", f"p12748-{n+1}.sto")
            with MSAFile(msa_path, digital=True, alphabet=self.alphabet) as msa_file:
                ref_msa = msa_file.read()

            self.assertEqual(len(iteration.msa), len(ref_msa))
            self.assertEqual(len(iteration.msa.sequences), len(ref_msa.sequences))
            self.assertEqual(iteration.msa.name, ref_msa.name)

            self.assertEqual(
                {seq.name for seq in iteration.msa.sequences},
                {seq.name for seq in ref_msa.sequences},
            )

        self.assertEqual(iteration.iteration, 3)
        self.assertTrue(iteration.converged)

    def test_iterate_hmm(self):
        # check that `Pipeline.iterate_hmm` produces consistent results
        # compared to a 'gold standard' run of `jackhmmer` (actually, manually iterated hmmsearches).

        hmm_path = resource_files(__package__).joinpath("data", "hmms", "txt", "KR.hmm")
        with HMMFile(hmm_path) as hmm_file:
            hmm = hmm_file.read()

        pipeline = Pipeline(alphabet=self.alphabet, incE=0.001, incdomE=0.001)
        search_iterator = pipeline.iterate_hmm(hmm, self.references)

        for n in range(6):
            self.assertEqual(search_iterator.iteration, n)
            iteration = next(search_iterator)

            self.assertEqual(search_iterator.iteration, n+1)
            self.assertEqual(iteration.iteration, n+1)

            msa_path = resource_files(__package__).joinpath("data", "jackhmmer", f"KR-{n+1}.sto")
            with MSAFile(msa_path, digital=True, alphabet=self.alphabet) as msa_file:
                ref_msa = msa_file.read()

            self.assertEqual(len(iteration.msa), len(ref_msa))
            self.assertEqual(len(iteration.msa.sequences), len(ref_msa.sequences))
            self.assertEqual(iteration.msa.name, ref_msa.name)

        self.assertEqual(iteration.iteration, 6)
        self.assertTrue(iteration.converged)


@unittest.skipUnless(resource_files, "importlib.resources.files not available")
class TestLongTargetsPipeline(unittest.TestCase):

    @staticmethod
    def _random_sequence(alphabet, name, length=5000):
        symbols = alphabet.symbols[:alphabet.K]
        return TextSequence(
            name=name,
            sequence="".join(random.sample(symbols, 1)[0] for i in range(length))
        ).digitize(alphabet)

    def test_search_hmm_block(self):
        dna = Alphabet.dna()
        rng = pyhmmer.easel.Randomness(0)

        targets = DigitalSequenceBlock(dna, [
            self._random_sequence(dna, b"seq1"),
            self._random_sequence(dna, b"seq2"),
        ])

        hmm = pyhmmer.plan7.HMM.sample(dna, 100, rng)
        hmm.name = b"test_one"
        hmm.accession = b"TST001"

        pipeline = LongTargetsPipeline(alphabet=dna)
        hits = pipeline.search_hmm(hmm, targets)
        self.assertEqual(hits.query.name, hmm.name)
        self.assertEqual(hits.query.accession, hmm.accession)

    def test_search_hmm_file(self):
        dna = Alphabet.dna()
        rng = pyhmmer.easel.Randomness(0)

        hmm = pyhmmer.plan7.HMM.sample(dna, 100, rng)
        hmm.name = b"test_one"
        hmm.accession = b"TST001"

        with tempfile.NamedTemporaryFile(mode="wb", suffix=".fna") as f:
            self._random_sequence(dna, b"seq1").write(f)
            self._random_sequence(dna, b"seq2").write(f)
            f.flush()

            with pyhmmer.easel.SequenceFile(f.name, digital=True, alphabet=dna) as targets:
                pipeline = LongTargetsPipeline(alphabet=dna)
                hits = pipeline.search_hmm(hmm, targets)

        self.assertEqual(hits.query.name, hmm.name)
        self.assertEqual(hits.query.accession, hmm.accession)

    def test_search_hmm_alphabet_mismatch(self):
        dna = Alphabet.dna()
        amino = Alphabet.amino()
        rng = pyhmmer.easel.Randomness(0)

        # mismatch between pipeline alphabet and database alphabet
        pipeline = LongTargetsPipeline(alphabet=dna)
        targets = DigitalSequenceBlock(amino)
        hmm = pyhmmer.plan7.HMM.sample(dna, 100, rng)
        self.assertRaises(AlphabetMismatch, pipeline.search_hmm, hmm, targets)

        # mismatch between pipeline alphabet and query alphabet
        pipeline = LongTargetsPipeline(alphabet=dna)
        targets = DigitalSequenceBlock(dna)
        hmm = pyhmmer.plan7.HMM.sample(amino, 100, rng)
        self.assertRaises(AlphabetMismatch, pipeline.search_hmm, hmm, targets)



from libc.stdint cimport int8_t, int16_t

cdef extern from "esl_matrixops.h" nogil:
    double **esl_mat_DCreate(int M, int N);
    float  **esl_mat_FCreate(int M, int N);
    int    **esl_mat_ICreate(int M, int N);
    char   **esl_mat_CCreate(int M, int N);

    double **esl_mat_DClone(double **A, int M, int N);
    float  **esl_mat_FClone(float **A,  int M, int N);
    int    **esl_mat_IClone(int **A,    int M, int N);

    int      esl_mat_DGrowTo(double ***ret_A, int M, int N);
    int      esl_mat_FGrowTo(float  ***ret_A, int M, int N);
    int      esl_mat_IGrowTo(int    ***ret_A, int M, int N);
    int      esl_mat_CGrowTo(char   ***ret_A, int M, int N);

    size_t   esl_mat_DSizeof(int M, int N);
    size_t   esl_mat_FSizeof(int M, int N);
    size_t   esl_mat_ISizeof(int M, int N);
    size_t   esl_mat_CSizeof(int M, int N);

    void     esl_mat_DSet(double **A, int M, int N, double value);
    void     esl_mat_FSet(float  **A, int M, int N, float  value);
    void     esl_mat_ISet(int    **A, int M, int N, int    value);

    void     esl_mat_DScale(double **A, int M, int N, double x);
    void     esl_mat_FScale(float **A,  int M, int N, float x);
    void     esl_mat_IScale(int **A, int M, int N, int x);

    void     esl_mat_DCopy(double  **src, int M, int N, double  **dest);
    void     esl_mat_FCopy(float   **src, int M, int N, float   **dest);
    void     esl_mat_ICopy(int     **src, int M, int N, int     **dest);
    void     esl_mat_WCopy(int16_t **src, int M, int N, int16_t **dest);
    void     esl_mat_BCopy(int8_t  **src, int M, int N, int8_t  **dest);

    double   esl_mat_DMax(double **A, int M, int N);
    float    esl_mat_FMax(float  **A, int M, int N);
    int      esl_mat_IMax(int    **A, int M, int N);

    int      esl_mat_DCompare(double **A, double **B, int M, int N, double tol);
    int      esl_mat_FCompare(float  **A, float  **B, int M, int N, float  tol);
    int      esl_mat_ICompare(int    **A, int    **B, int M, int N);

    void     esl_mat_DDestroy(double **A);
    void     esl_mat_FDestroy(float  **A);
    void     esl_mat_IDestroy(int    **A);
    void     esl_mat_CDestroy(char   **A);

    int      esl_mat_DDump(double **A, int M, int N);
    int      esl_mat_FDump( float **A, int M, int N);
    int      esl_mat_IDump(   int **A, int M, int N);



cdef extern from "esl_cluster.h" nogil:

    int esl_cluster_SingleLinkage(
        void *base, size_t n, size_t size,
        int (*linkfunc)(const void *, const void *, const void *, int *),
        void *param,
        int *workspace, int *assignments, int *ret_C
    );



from libc.stdint cimport uint8_t, uint32_t, int64_t

from libeasel cimport eslERRBUFSIZE
from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.dmatrix cimport ESL_DMATRIX
from libeasel.getopts cimport ESL_GETOPTS
from libeasel.msa cimport ESL_MSA
from libeasel.random cimport ESL_RANDOMNESS
from libeasel.scorematrix cimport ESL_SCOREMATRIX
from libeasel.sq cimport ESL_SQ
from libhmmer.p7_bg cimport P7_BG
from libhmmer.p7_hmm cimport P7_HMM
from libhmmer.p7_prior cimport P7_PRIOR
from libhmmer.p7_profile cimport P7_PROFILE
from libhmmer.p7_trace cimport P7_TRACE

if HMMER_IMPL == "VMX":
    from libhmmer.impl_vmx.p7_oprofile cimport P7_OPROFILE
elif HMMER_IMPL == "SSE":
    from libhmmer.impl_sse.p7_oprofile cimport P7_OPROFILE
elif HMMER_IMPL == "NEON":
    from libhmmer.impl_neon.p7_oprofile cimport P7_OPROFILE


cdef extern from "hmmer.h" nogil:

    const double p7_DEFAULT_WINDOW_BETA

    cdef enum p7_archchoice_e:
        p7_ARCH_FAST = 0
        p7_ARCH_HAND = 1

    cdef enum p7_wgtchoice_e:
        p7_WGT_NONE   = 0
        p7_WGT_GIVEN  = 1
        p7_WGT_GSC    = 2
        p7_WGT_PB     = 3
        p7_WGT_BLOSUM = 4

    cdef enum p7_effnchoice_e:
        p7_EFFN_NONE        = 0
        p7_EFFN_SET         = 1
        p7_EFFN_CLUST       = 2
        p7_EFFN_ENTROPY     = 3
        p7_EFFN_ENTROPY_EXP = 4

    ctypedef p7_builder_s P7_BUILDER
    cdef struct p7_builder_s:
        p7_archchoice_e arch_strategy
        float symfrac
        float fragthresh

        p7_wgtchoice_e wgt_strategy
        double wid

        p7_effnchoice_e effn_strategy
        double re_target
        double esigma
        double eid
        double eset

        ESL_RANDOMNESS* r
        int do_reseeding

        int EmL
        int EmN
        int EvL
        int EvN
        int EfL
        int EfN
        double Eft

        P7_PRIOR* prior
        int max_insert_len

        ESL_SCOREMATRIX* S
        ESL_DMATRIX* Q
        double popen
        double pextend

        double w_beta
        int w_len

        const ESL_ALPHABET* abc
        char errbuf[eslERRBUFSIZE]

    P7_BUILDER *p7_builder_Create(const ESL_GETOPTS *go, const ESL_ALPHABET *abc)
    int         p7_builder_LoadScoreSystem(P7_BUILDER *bld, const char *matrix,                  double popen, double pextend, P7_BG *bg)
    int         p7_builder_SetScoreSystem (P7_BUILDER *bld, const char *mxfile, const char *env, double popen, double pextend, P7_BG *bg)
    void        p7_builder_Destroy(P7_BUILDER *bld)

    int p7_Builder      (P7_BUILDER *bld, ESL_MSA *msa, P7_BG *bg, P7_HMM **opt_hmm, P7_TRACE ***opt_trarr, P7_PROFILE **opt_gm, P7_OPROFILE **opt_om, ESL_MSA **opt_postmsa)
    int p7_SingleBuilder(P7_BUILDER *bld, ESL_SQ *sq,   P7_BG *bg, P7_HMM **opt_hmm, P7_TRACE  **opt_tr,    P7_PROFILE **opt_gm, P7_OPROFILE **opt_om)
    int p7_Builder_MaxLength      (P7_HMM *hmm, double emit_thresh)



from libc.stdint cimport uint8_t, int64_t


cdef extern from "stdarg.h" nogil:
    ctypedef struct va_list:
          pass

cdef extern from "easel.h" nogil:

    const size_t eslERRBUFSIZE

    const double eslCONST_E
    const double eslCONST_PI
    const double eslCONST_EULER
    const double eslCONST_GOLD
    const double eslCONST_LOG2
    const double eslCONST_LOG2R
    const double eslINFINITY

    ctypedef uint8_t ESL_DSQ
    const    ESL_DSQ eslDSQ_SENTINEL
    const    ESL_DSQ eslDSQ_ILLEGAL
    const    ESL_DSQ eslDSQ_IGNORED
    const    ESL_DSQ eslDSQ_EOL
    const    ESL_DSQ eslDSQ_EOD

    ctypedef int64_t esl_pos_t

    cdef enum:
        eslOK = 0
        eslFAIL = 1
        eslEOL = 2
        eslEOF = 3
        eslEOD = 4
        eslEMEM = 5
        eslENOTFOUND = 6
        eslEFORMAT = 7
        eslEAMBIGUOUS = 8
        eslEDIVZERO = 9
        eslEINCOMPAT = 10
        eslEINVAL = 11
        eslESYS = 12
        eslECORRUPT = 13
        eslEINCONCEIVABLE = 14
        eslESYNTAX = 15
        eslERANGE = 16
        eslEDUP = 17
        eslENOHALT = 18
        eslENORESULT = 19
        eslENODATA = 20
        eslETYPE = 21
        eslEOVERWRITE = 22
        eslENOSPACE = 23
        eslEUNIMPLEMENTED = 24
        eslENOFORMAT = 25
        eslENOALPHABET = 26
        eslEWRITE = 27
        eslEINACCURATE = 28

    # 1. Error handling
    ctypedef void(*esl_exception_handler_f)(int errcode, int use_errno, char *sourcefile, int sourceline, char *format, va_list argp)
    # void esl_fail(char *errbuf, const char *format, ...)
    # void esl_exception(int errcode, int use_errno, char *sourcefile, int sourceline, char *format, ...);
    void esl_exception_SetHandler(esl_exception_handler_f)
    void esl_exception_ResetDefaultHandler()
    # void esl_nonfatal_handler(int errcode, int use_errno, char *sourcefile, int sourceline, char *format, va_list argp);
    # void esl_fatal(const char *format, ...) ESL_ATTRIBUTE_NORETURN;

    # 4. Improved replacements for some C library functions
    # extern int  esl_fgets(char **buf, int *n, FILE *fp);
    # extern int  esl_fprintf(FILE *fp, const char *format, ...);
    # extern int  esl_printf(const char *format, ...);
    # extern int  esl_strdup(const char *s, int64_t n, char **ret_dup);
    # extern int  esl_strcat(char **dest, int64_t ldest, const char *src, int64_t lsrc);
    # extern int  esl_strmapcat        (const ESL_DSQ *inmap, char **dest, int64_t *ldest, const char *src, esl_pos_t lsrc);
    # extern int  esl_strmapcat_noalloc(const ESL_DSQ *inmap,  char *dest, int64_t *ldest, const char *src, esl_pos_t lsrc);
    # extern int  esl_strtok    (char **s, char *delim, char **ret_tok);
    # extern int  esl_strtok_adv(char **s, char *delim, char **ret_tok, int *opt_toklen, char *opt_endchar);
    # extern int  esl_sprintf (char **ret_s, const char *format, ...);
    # extern int  esl_vsprintf(char **ret_s, const char *format, va_list *ap);
    # extern int  esl_strcmp(const char *s1, const char *s2);

    # 5.  Portable drop-in replacements for non-standard C functions
    extern int  esl_strcasecmp(const char *s1, const char *s2)
    extern char *esl_strsep(char **stringp, const char *delim)

    # 6. Additional string functions, esl_str*()
    # int     esl_strchop(char *s, int64_t n)
    # int     esl_strdealign(char *s, const char *aseq, const char *gapchars, int64_t *opt_rlen)
    # int     esl_str_IsBlank(char *s)
    # int     esl_str_IsInteger(char *s)
    # int     esl_str_IsReal(char *s)
    # int64_t esl_str_GetMaxWidth(char **s, int n)

    # 7. File path/name manipulation functions, including tmpfiles
    # int  esl_FileExists(const char *filename)
    # int  esl_FileTail(const char *path, int nosuffix, char **ret_file)
    # int  esl_file_Extension(char *filename, esl_pos_t n_ignore, char **ret_sfx, esl_pos_t *ret_n)
    # int  esl_FileConcat(const char *dir, const char *file, char **ret_path)
    # int  esl_FileNewSuffix(const char *filename, const char *sfx, char **ret_newpath)
    # int  esl_FileEnvOpen(const char *fname, const char *env, FILE **ret_fp, char **ret_path);
    # int  esl_tmpfile(char *basename6X, FILE **ret_fp)
    # int  esl_tmpfile_named(char *basename6X, FILE **ret_fp)
    # int  esl_getcwd(char **ret_cwd)

    # 8. Typed comparison routines.
    int esl_DCompare(double x0, double x, double r_tol, double a_tol)
    int esl_FCompare(float  x0, float  x, float  r_tol, float  a_tol)
    int esl_CCompare(char *s1, char *s2)
    int esl_DCompare_old(double a,  double b, double tol)
    int esl_FCompare_old(float  a,  float  b, float  tol)


cdef extern from "hmmer.h" nogil:

    ctypedef struct p7_spcoord_s:
        int idx
        int i
        int j
        int k
        int m
        float prob

    ctypedef p7_spensemble_s P7_SPENSEMBLE
    cdef struct p7_spensemble_s:
        int nsamples
        p7_spcoord_s* sp
        int nalloc
        int n

        int* workspace
        int* assignment
        int nc

        int* epc
        int epc_alloc

        p7_spcoord_s* sigc
        int nsigc
        int nsigc_alloc


    P7_SPENSEMBLE *p7_spensemble_Create(int init_n, int init_epc, int init_sigc)
    int            p7_spensemble_Reuse(P7_SPENSEMBLE *sp)
    int            p7_spensemble_Add(P7_SPENSEMBLE *sp, int sampleidx, int i, int j, int k, int m)
    int            p7_spensemble_Cluster(P7_SPENSEMBLE *sp,
    				          float min_overlap, int of_smaller, int max_diagdiff,
    				          float min_posterior, float min_endpointp,
    				          int *ret_nclusters)
    int     p7_spensemble_GetClusterCoords(P7_SPENSEMBLE *sp, int which,
    					        int *ret_i, int *ret_j, int *ret_k, int *ret_m, float *ret_p)
    void    p7_spensemble_Destroy(P7_SPENSEMBLE *sp)



import copy
import functools
import gc
import io
import os
import unittest
import tempfile

from pyhmmer import easel

from .. import __name__ as __package__
from .utils import EASEL_FOLDER, resource_files


class _TestMSA(object):
    @classmethod
    def setUpClass(cls):
        cls.formats_folder = os.path.join(EASEL_FOLDER, "formats")

    def assertAlmostEqual(self, x, y, places=None):
        if isinstance(x, easel.Vector) and isinstance(y, easel.Vector):
            diff = abs(y - x).sum()
            super().assertAlmostEqual(diff, 0, msg=f'{x} != {y} within {places} places ({diff} difference)', places=places)
        else:
            super().assertAlmostEqual(x, y, places=places)

    @unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
    def test_write_roundtrip_stockholm(self):
        sto = os.path.join(self.formats_folder, "stockholm.1")
        msa = self.read_msa(sto)
        with io.BytesIO() as buffer:
            msa.write(buffer, "stockholm")
            actual = buffer.getvalue().decode()
        with open(sto) as f:
            expected = f.read()
        self.assertMultiLineEqual(actual, expected)

    def test_write_invalid_format(self):
        msa = easel.TextMSA()
        self.assertRaises(ValueError, msa.write, io.BytesIO(), "invalidformat")

    @unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
    def test_sequences(self):
        sto = os.path.join(self.formats_folder, "stockholm.1")
        msa = self.read_msa(sto)
        self.assertEqual(len(msa.sequences), 2)
        self.assertEqual(msa.sequences[0].name, b"seq1")
        self.assertEqual(msa.sequences[1].name, b"seq2")
        self.assertEqual(msa.sequences[-2].name, b"seq1")
        self.assertEqual(msa.sequences[-1].name, b"seq2")

        with self.assertRaises(IndexError):
            msa.sequences[-3]

    def test_init_empty(self):
        msa = self.MSA()
        self.assertEqual(len(msa), 0)
        self.assertEqual(len(msa.sequences), 0)
        self.assertFalse(msa)

    def test_init_name(self):
        msa = self.MSA(name=b"ali")
        self.assertEqual(msa.name, b"ali")

    def test_init_description(self):
        d = b"an alignment made from Python"
        msa = self.MSA(description=d)
        self.assertEqual(msa.description, d)

    def test_init_author(self):
        author = b"Martin Larralde"
        msa = self.MSA(author=author)
        self.assertEqual(msa.author, author)

    def test_init_accession(self):
        acc = b"TST001"
        msa = self.MSA(accession=acc)
        self.assertEqual(msa.accession, acc)

    def test_indexed_key_error(self):
        msa = self.MSA()
        with self.assertRaises(KeyError):
            msa.indexed[b"xxx"]

    def test_indexed_type_error(self):
        msa = self.MSA()
        with self.assertRaises(TypeError):
            msa.indexed[1]

    def test_mark_fragments(self):
        dna = easel.Alphabet.dna()
        buffer = io.BytesIO(
            b"# STOCKHOLM 1.0\n\n"
            b"seq1  ..AAAAAA..\n"
            b"seq2  .....AAAA.\n"
            b"seq3  .CCC.CCCCC\n"
            b"seq4  GGGGGGGGGG\n"
            b"//\n"
        )
        msa = self.read_msa(buffer, alphabet=dna)

        f1 = msa.mark_fragments(0.5)    
        self.assertEqual(len(f1), 4)
        self.assertFalse(f1[0])
        self.assertTrue(f1[1])
        self.assertFalse(f1[2])
        self.assertFalse(f1[3])

        f2 = msa.mark_fragments(0.95)    
        self.assertEqual(len(f2), 4)
        self.assertTrue(f2[0])
        self.assertTrue(f2[1])
        self.assertTrue(f2[2])
        self.assertFalse(f2[3])

    def test_mark_fragments_error(self):
        msa = self.MSA()
        with self.assertRaises(ValueError):
            _ = msa.mark_fragments(200.0)

    def test_compute_weights_identical_seqs_nt(self):
        nt = easel.Alphabet.dna()
        buffer = io.BytesIO(b"# STOCKHOLM 1.0\n\nseq1 AAAAA\nseq2 AAAAA\nseq3 AAAAA\nseq4 AAAAA\nseq5 AAAAA\n//\n")
        msa = self.read_msa(buffer, alphabet=nt)
        uniform = easel.VectorD([1.0, 1.0, 1.0, 1.0, 1.0])
        for method in ["gsc", "pb", "blosum"]:
            self.assertEqual(msa.compute_weights(method), uniform)

    def test_compute_weights_identical_seqs_aa(self):
        aa = easel.Alphabet.amino()
        buffer = io.BytesIO(b"# STOCKHOLM 1.0\n\nseq1 AAAAA\nseq2 AAAAA\nseq3 AAAAA\nseq4 AAAAA\nseq5 AAAAA\n//\n")
        msa = self.read_msa(buffer, alphabet=aa)
        uniform = easel.VectorD([1.0, 1.0, 1.0, 1.0, 1.0])
        for method in ["gsc", "pb", "blosum"]:
            self.assertEqual(msa.compute_weights(method), uniform)

    def test_compute_weights_henikoff_contrived(self):
        aa = easel.Alphabet.amino()
        buffer = io.BytesIO(b"# STOCKHOLM 1.0\n\nseq1 AAAAA\nseq2 AAAAA\nseq3 CCCCC\nseq4 CCCCC\nseq5 TTTTT\n//\n")
        msa = self.read_msa(buffer, alphabet=aa)
        expected = easel.VectorD([0.833333, 0.833333, 0.833333, 0.833333, 1.66667])
        for method in ["gsc", "pb", "blosum"]:
            self.assertAlmostEqual(msa.compute_weights(method), expected, places=3)

    def test_compute_weights_nitrogenase(self):
        aa = easel.Alphabet.amino()
        buffer = io.BytesIO(b"# STOCKHOLM 1.0\n\nNIFE_CLOPA GYVGS\nNIFD_AZOVI GFDGF\nNIFD_BRAJA GYDGF\nNIFK_ANASP GYQGG\n//\n")
        msa = self.read_msa(buffer, alphabet=aa)

        exp_gsc = easel.VectorD([1.125000, 0.875000, 0.875000, 1.125000])
        exp_pb  = easel.VectorD([1.066667, 1.066667, 0.800000, 1.066667])
        exp_blo = easel.VectorD([1.333333, 0.666667, 0.666667, 1.333333])

        self.assertAlmostEqual(msa.compute_weights("gsc"), exp_gsc, places=3)
        self.assertAlmostEqual(msa.compute_weights("pb"), exp_pb, places=3)
        self.assertAlmostEqual(msa.compute_weights("blosum"), exp_blo, places=3)


class TestTextMSA(_TestMSA, unittest.TestCase):

    MSA = easel.TextMSA

    @staticmethod
    def read_msa(sto, alphabet=None):
        with easel.MSAFile(sto, "stockholm", alphabet=alphabet) as msa_file:
            return msa_file.read()

    def test_indexed(self):
        s1 = easel.TextSequence(name=b"seq1", sequence="ATGC")
        s2 = easel.TextSequence(name=b"seq2", sequence="ATGG")
        msa = easel.TextMSA(sequences=[s1, s2])
        self.assertEqual(len(msa.indexed), 2)
        self.assertEqual(msa.indexed[b"seq1"], s1)
        self.assertEqual(msa.indexed[b"seq2"], s2)

    def test_rf(self):
        s1 = easel.TextSequence(name=b"seq1", sequence="ATGC")
        s2 = easel.TextSequence(name=b"seq2", sequence="ATGG")
        msa = easel.TextMSA(sequences=[s1, s2])
        self.assertIsNone(msa.reference)
        msa.reference = b"xxxx"
        self.assertEqual(msa.reference, b"xxxx")
        with self.assertRaises(ValueError):
            msa.reference = b"x"

    def test_eq(self):
        s1 = easel.TextSequence(name=b"seq1", sequence="ATGC")
        s2 = easel.TextSequence(name=b"seq2", sequence="ATGG")
        msa = easel.TextMSA(sequences=[s1, s2])
        msa2 = easel.TextMSA(sequences=[s1, s2])
        self.assertEqual(msa, msa2)

        msa2.name = b"other"
        self.assertNotEqual(msa, msa2)

        self.assertNotEqual(msa, 1)
        self.assertNotEqual(msa, [s1, s2])

    def test_eq_copy(self):
        s1 = easel.TextSequence(name=b"seq1", sequence="ATGC")
        s2 = easel.TextSequence(name=b"seq2", sequence="ATGG")
        msa = easel.TextMSA(sequences=[s1, s2])
        self.assertEqual(msa, msa.copy())
        self.assertEqual(msa, copy.copy(msa))

    def test_init_error(self):
        self.assertRaises(TypeError, easel.TextMSA, sequences=[1, 2, 3])

    def test_init_sequences(self):
        s1 = easel.TextSequence(name=b"seq1", sequence="ATGC")
        msa = easel.TextMSA(sequences=[s1])
        self.assertEqual(len(msa), 4)
        self.assertEqual(len(msa.sequences), 1)
        self.assertTrue(msa)

    def test_init_length_mismatch(self):
        s1 = easel.TextSequence(name=b"seq1", sequence="ATGC")
        s2 = easel.TextSequence(name=b"seq2", sequence="AT")
        self.assertRaises(ValueError, easel.TextMSA, sequences=[s1, s2])

    def test_init_duplicate_names(self):
        s1 = easel.TextSequence(name=b"seq1", sequence="ATGC")
        s2 = easel.TextSequence(name=b"seq1", sequence="ATTC")
        self.assertRaises(ValueError, easel.TextMSA, sequences=[s1, s2])

    def test_insert_sequences(self):
        s1 = easel.TextSequence(name=b"seq1", sequence="ATGC", description=b"first", accession=b"SQ.1")
        s2 = easel.TextSequence(name=b"seq2", sequence="ATTC", description=b"second", accession=b"SQ.2")
        msa = easel.TextMSA(sequences=[s1, s2])
        self.assertEqual(len(msa), 4)
        self.assertEqual(len(msa.sequences), 2)
        self.assertEqual(msa.sequences[0].name, b"seq1")
        self.assertEqual(msa.sequences[1].name, b"seq2")

        s3 = easel.TextSequence(name=b"seq3", sequence="AGGC")
        msa.sequences[0] = s3
        self.assertEqual(len(msa), 4)
        self.assertEqual(len(msa.sequences), 2)
        self.assertEqual(msa.sequences[0].name, b"seq3")
        self.assertEqual(msa.sequences[1].name, b"seq2")

        s4 = easel.TextSequence(name=b"seq4", sequence="TTTT")
        msa.sequences[-1] = s4
        self.assertEqual(len(msa), 4)
        self.assertEqual(len(msa.sequences), 2)
        self.assertEqual(msa.sequences[0].name, b"seq3")
        self.assertEqual(msa.sequences[1].name, b"seq4")

        s5 = easel.TextSequence(name=b"seq5", sequence="AAAA")
        with self.assertRaises(IndexError):
            msa.sequences[2] = s5
        with self.assertRaises(IndexError):
            msa.sequences[-10] = s5

        s4bis = easel.TextSequence(name=b"seq4", sequence="TTTT")
        with self.assertRaises(ValueError):
            msa.sequences[0] = s4bis

        s6 = easel.TextSequence(name=b"seq4", sequence="TT")
        with self.assertRaises(ValueError):
            msa.sequences[0] = s6

        dna = easel.Alphabet.dna()
        s7 = easel.DigitalSequence(dna, name=b"seq4", sequence=bytearray([2, 2]))
        with self.assertRaises(TypeError):
            msa.sequences[0] = s7

    def test_digitize(self):
        dna = easel.Alphabet.dna()
        s1 = easel.TextSequence(name=b"seq1", sequence="ACGT")
        msa_t = easel.TextMSA(sequences=[s1])
        msa_d = msa_t.digitize(dna)

        self.assertIsInstance(msa_d, easel.DigitalMSA)
        self.assertEqual(len(msa_d), 4)
        self.assertEqual(len(msa_d.sequences), 1)
        self.assertTrue(msa_d)
        self.assertEqual(msa_d.sequences[0].name, b"seq1")
        self.assertEqual(msa_d.sequences[0].sequence, easel.VectorU8([0, 1, 2, 3]))

    def test_digitize_invalid(self):
        dna = easel.Alphabet.dna()
        s1 = easel.TextSequence(name=b"seq1", sequence=">v3ry 1nv4l1d")
        msa_t = easel.TextMSA(sequences=[s1])
        self.assertRaises(ValueError, msa_t.digitize, dna)


class TestDigitalMSA(_TestMSA, unittest.TestCase):

    alphabet = easel.Alphabet.dna()
    MSA = staticmethod(functools.partial(easel.DigitalMSA, alphabet))

    @staticmethod
    def read_msa(sto, alphabet=None):
        with easel.MSAFile(sto, "stockholm", digital=True, alphabet=alphabet) as msa_file:
            return msa_file.read()

    def test_sample(self):
        rng = easel.Randomness(42)
        msa = easel.DigitalMSA.sample(self.alphabet, 10, 50, rng)
        self.assertLessEqual(len(msa.sequences), 10)
        self.assertLessEqual(len(msa), 50)

    def test_sample_seed(self):
        msa1 = easel.DigitalMSA.sample(self.alphabet, 10, 50, randomness=easel.Randomness(42))
        msa2 = easel.DigitalMSA.sample(self.alphabet, 10, 50, randomness=42)
        self.assertEqual(msa1, msa2)
        msa3 = easel.DigitalMSA.sample(self.alphabet, 10, 50, randomness=100)
        self.assertNotEqual(msa3, msa2)

        with self.assertRaises(TypeError):
            msa4 = easel.DigitalMSA.sample(self.alphabet, 10, 50, randomness="xxx")

    def test_eq(self):
        s1 = easel.DigitalSequence(self.alphabet, name=b"seq1", sequence=bytearray([1, 2, 3, 4]))
        s2 = easel.DigitalSequence(self.alphabet, name=b"seq2", sequence=bytearray([1, 2, 3, 3]))
        msa = easel.DigitalMSA(self.alphabet, sequences=[s1, s2])
        msa2 = easel.DigitalMSA(self.alphabet, sequences=[s1, s2])
        self.assertEqual(msa, msa2)

        msa2.name = b"other"
        self.assertNotEqual(msa, msa2)

        self.assertNotEqual(msa, 1)
        self.assertNotEqual(msa, [s1, s2])

    def test_eq_copy(self):
        s1 = easel.DigitalSequence(self.alphabet, name=b"seq1", sequence=bytearray([1, 2, 3, 4]))
        s2 = easel.DigitalSequence(self.alphabet, name=b"seq2", sequence=bytearray([1, 2, 3, 3]))
        msa = easel.DigitalMSA(self.alphabet, sequences=[s1, s2])
        self.assertEqual(msa, msa.copy())
        self.assertEqual(msa, copy.copy(msa))

    def test_init_error(self):
        self.assertRaises(TypeError, easel.TextMSA, sequences=[1, 2, 3])

    def test_init_sequences(self):
        s1 = easel.DigitalSequence(self.alphabet, name=b"seq1", sequence=bytearray([1, 2, 3, 4]))
        msa = easel.DigitalMSA(self.alphabet, sequences=[s1])
        self.assertEqual(len(msa), 4)
        self.assertEqual(len(msa.sequences), 1)
        self.assertTrue(msa)

    def test_init_length_mismatch(self):
        s1 = easel.DigitalSequence(self.alphabet, name=b"seq1", sequence=bytearray([1, 2, 3, 4]))
        s2 = easel.DigitalSequence(self.alphabet, name=b"seq2", sequence=bytearray([1, 2]))
        self.assertRaises(ValueError, easel.DigitalMSA, self.alphabet, sequences=[s1, s2])

    def test_init_duplicate_names(self):
        s1 = easel.DigitalSequence(self.alphabet, name=b"seq1", sequence=bytearray([1, 2, 3, 4]))
        s2 = easel.DigitalSequence(self.alphabet, name=b"seq1", sequence=bytearray([1, 2, 3, 3]))
        self.assertRaises(ValueError, easel.DigitalMSA, self.alphabet, sequences=[s1, s2])

    def test_insert_sequences(self):
        s1 = easel.DigitalSequence(self.alphabet, name=b"seq1", sequence=bytearray([1, 2, 3, 4]), description=b"first", accession=b"SQ.1")
        s2 = easel.DigitalSequence(self.alphabet, name=b"seq2", sequence=bytearray([1, 3, 3, 4]), description=b"second", accession=b"SQ.2")
        msa = easel.DigitalMSA(self.alphabet, sequences=[s1, s2])
        self.assertEqual(len(msa), 4)
        self.assertEqual(len(msa.sequences), 2)
        self.assertEqual(msa.sequences[0].name, b"seq1")
        self.assertEqual(msa.sequences[1].name, b"seq2")

        s3 = easel.DigitalSequence(self.alphabet, name=b"seq3", sequence=bytearray([4, 2, 3, 4]))
        msa.sequences[0] = s3
        self.assertEqual(len(msa), 4)
        self.assertEqual(len(msa.sequences), 2)
        self.assertEqual(msa.sequences[0].name, b"seq3")
        self.assertEqual(msa.sequences[1].name, b"seq2")

        s4 = easel.DigitalSequence(self.alphabet, name=b"seq4", sequence=bytearray([2, 2, 2, 2]))
        msa.sequences[-1] = s4
        self.assertEqual(len(msa), 4)
        self.assertEqual(len(msa.sequences), 2)
        self.assertEqual(msa.sequences[0].name, b"seq3")
        self.assertEqual(msa.sequences[1].name, b"seq4")

        s5 = easel.DigitalSequence(self.alphabet, name=b"seq5", sequence=bytearray([1, 1, 1, 1]))
        with self.assertRaises(IndexError):
            msa.sequences[2] = s5
        with self.assertRaises(IndexError):
            msa.sequences[-10] = s5

        s4bis = easel.DigitalSequence(self.alphabet, name=b"seq4", sequence=bytearray([2, 2, 2, 2]))
        with self.assertRaises(ValueError):
            msa.sequences[0] = s4bis

        s6 = easel.DigitalSequence(self.alphabet, name=b"seq4", sequence=bytearray([2, 2]))
        with self.assertRaises(ValueError):
            msa.sequences[0] = s6

        s7 = easel.TextSequence(name=b"seq4", sequence="ATGC")
        with self.assertRaises(TypeError):
            msa.sequences[0] = s7

    def test_textize(self):
        dna = easel.Alphabet.dna()
        s1 = easel.DigitalSequence(self.alphabet, name=b"seq1", sequence=bytearray([0, 1, 2, 3]))
        msa_d = easel.DigitalMSA(self.alphabet, sequences=[s1])
        msa_t = msa_d.textize()

        self.assertIsInstance(msa_t, easel.TextMSA)
        self.assertEqual(len(msa_t), 4)
        self.assertEqual(len(msa_t.sequences), 1)
        self.assertTrue(msa_t)
        self.assertEqual(msa_t.sequences[0].name, b"seq1")
        self.assertEqual(msa_t.sequences[0].sequence, "ACGT")

    def test_identity_filter(self):
        # adapted from `utest_idfilter` in `esl_msaweight.c`
        dna = easel.Alphabet.dna()
        s1 = easel.DigitalSequence

        buffer = io.BytesIO(
            b"# STOCKHOLM 1.0\n\n"
            b"seq1  ..AAAAAAAA\n"
            b"seq2  AAAAAAAAAA\n"
            b"seq3  CCCCCCCCCC\n"
            b"seq4  GGGGGGGGGG\n"
            b"//\n"
        )

        with easel.MSAFile(buffer, format="stockholm", digital=True, alphabet=dna) as msa_file:
            msa = msa_file.read()

        msa2 = msa.identity_filter(1.0)
        self.assertEqual(len(msa2.sequences), 3)
        self.assertEqual(msa2.names[0], b"seq2")

        msa3 = msa.identity_filter(1.0, preference="origorder")
        self.assertEqual(len(msa3.sequences), 3)
        self.assertEqual(msa3.names[0], b"seq1")

        msa3 = msa.identity_filter(1.0, preference="random")
        self.assertEqual(len(msa3.sequences), 3)
        self.assertIn(msa3.names[0], [b"seq1", b"seq2"])

    @unittest.skipUnless(resource_files, "importlib.resources.files not available")
    def test_identity_filter_luxc(self):
        luxc = resource_files(__package__).joinpath("data", "msa", "LuxC.sto")
        msa = self.read_msa(luxc)
        filtered = msa.identity_filter()

        weighted = resource_files(__package__).joinpath("data", "msa", "LuxC.weighted.sto")
        expected = self.read_msa(weighted)

        self.assertEqual(len(filtered.sequences), len(expected.sequences))
        self.assertEqual(len(filtered.names), len(expected.names))
        for s1, s2 in zip(filtered.sequences, expected.sequences):
            self.assertEqual(s1, s2)

    def test_reverse_complement(self):
        dna = easel.Alphabet.dna()

        buffer = io.BytesIO(
            b"# STOCKHOLM 1.0\n\n"
            b"seq1  ..TTAAAACC\n"
            b"seq2  AAAAGGAATT\n"
            b"seq3  CCCCGGGG..\n"
            b"seq4  GGGGGGGGGG\n"
            b"//\n"
        )

        with easel.MSAFile(buffer, format="stockholm", digital=True, alphabet=dna) as msa_file:
            msa = msa_file.read()

        rc = msa.reverse_complement()
        self.assertEqual(dna.decode(rc.alignment[0]), "GGTTTTAA--")
        self.assertEqual(dna.decode(rc.alignment[1]), "AATTCCTTTT")
        self.assertEqual(dna.decode(rc.alignment[2]), "--CCCCGGGG")
        self.assertEqual(dna.decode(rc.alignment[3]), "CCCCCCCCCC")
        
        msa.reverse_complement(inplace=True)
        self.assertEqual(dna.decode(msa.alignment[0]), "GGTTTTAA--")
        self.assertEqual(dna.decode(msa.alignment[1]), "AATTCCTTTT")
        self.assertEqual(dna.decode(msa.alignment[2]), "--CCCCGGGG")
        self.assertEqual(dna.decode(msa.alignment[3]), "CCCCCCCCCC")


import collections
import copy
import queue
import os
import itertools
import multiprocessing
import sys
import typing
import threading

import psutil

from ..easel import Alphabet, DigitalSequence, DigitalMSA, DigitalSequenceBlock, SequenceFile
from ..plan7 import TopHits, HMM, Profile, OptimizedProfile, Pipeline, Builder, IterationResult
from ..utils import singledispatchmethod, peekable
from ._base import _BaseDispatcher, _BaseWorker, _BaseChore, _AnyProfile

_JACKHMMERQueryType = typing.Union[DigitalSequence, _AnyProfile]
_I = typing.TypeVar("_I") # generic iteration result

if typing.TYPE_CHECKING:

    if sys.version_info >= (3, 8):
        from typing import Literal
    else:
        from typing_extensions import Literal  # type: ignore

    if sys.version_info >= (3, 11):
        from typing import Unpack
    else:
        from typing_extensions import Unpack

    from ._base import PipelineOptions, BACKEND

# --- Worker -------------------------------------------------------------------

class _JACKHMMERWorker(
    typing.Generic[_I],
    _BaseWorker[
        _JACKHMMERQueryType,
        DigitalSequenceBlock,
        _I,
    ],
):
    def __init__(
        self,
        targets: DigitalSequenceBlock,
        query_queue: "queue.Queue[typing.Optional[_BaseChore[_JACKHMMERQueryType, _I]]]",
        query_count: multiprocessing.Value,  # type: ignore
        kill_switch: threading.Event,
        callback: typing.Optional[typing.Callable[[_JACKHMMERQueryType, int], None]],
        options: "PipelineOptions",
        builder: typing.Optional[Builder] = None,
        max_iterations: typing.Optional[int] = 5,
        select_hits: typing.Optional[typing.Callable[["TopHits[_JACKHMMERQueryType]"], None]] = None,
        checkpoints: bool = False,
    ) -> None:
        super().__init__(
            targets=targets,
            query_queue=query_queue,
            query_count=query_count,
            kill_switch=kill_switch,
            callback=callback,
            options=options,
            builder=builder,
        )
        self.select_hits = select_hits
        self.max_iterations = max_iterations
        self.checkpoints = checkpoints

    @singledispatchmethod
    def query(self, query) -> typing.Union[IterationResult, typing.Iterable[IterationResult]]:  # type: ignore
        raise TypeError(
            "Unsupported query type for `jackhmmer`: {}".format(type(query).__name__)
        )

    @query.register(DigitalSequence)
    def _(self, query: DigitalSequence) -> typing.Union[IterationResult, typing.Iterable[IterationResult]]:  # type: ignore
        assert self.pipeline is not None
        iterator = self.pipeline.iterate_seq(
            query,
            self.targets,
            self.builder,
            self.select_hits  # type: ignore
        )
        return self._iterate(iterator, self.checkpoints)

    @query.register(HMM)
    def _(self, query: HMM) -> typing.Union[IterationResult, typing.Iterable[IterationResult]]:  # type: ignore
        assert self.pipeline is not None
        iterator = self.pipeline.iterate_hmm(
            query,
            self.targets,
            self.builder,
            self.select_hits  # type: ignore
        )
        return self._iterate(iterator, self.checkpoints)

    @typing.overload
    def _iterate(
        self, iterator: typing.Iterable[IterationResult], checkpoints: "Literal[False]"
    ) -> IterationResult:
        ...

    @typing.overload
    def _iterate(
        self, iterator: typing.Iterable[IterationResult], checkpoints: "Literal[True]"
    ) -> typing.Iterable[IterationResult]:
        ...

    @typing.overload
    def _iterate(
        self, iterator: typing.Iterable[IterationResult], checkpoints: bool = False
    ) -> typing.Union[IterationResult, typing.Iterable[IterationResult]]:
        ...

    def _iterate(
        self, iterator: typing.Iterable[IterationResult], checkpoints: bool = False
    ) -> typing.Union[IterationResult, typing.Iterable[IterationResult]]:
        iteration_checkpoints = []
        for iteration in itertools.islice(iterator, self.max_iterations):
            if checkpoints:
                iteration_checkpoints.append(iteration)
            if iteration.converged:
                break
        return iteration_checkpoints if checkpoints else iteration


class _JACKHMMERThread(_JACKHMMERWorker, threading.Thread):
    pass


class _JACKHMMERProcess(_JACKHMMERWorker, multiprocessing.Process):
    pass


# --- Dispatcher ---------------------------------------------------------------

class _JACKHMMERDispatcher(
    typing.Generic[_I],
    _BaseDispatcher[
        _JACKHMMERQueryType,
        DigitalSequenceBlock,
        _I,
    ],
):
    """A dispatcher to run JackHMMER iterative searches."""

    def __init__(
        self,
        queries: typing.Iterable[_JACKHMMERQueryType],
        targets: DigitalSequenceBlock,
        cpus: int = 0,
        callback: typing.Optional[
            typing.Callable[[_JACKHMMERQueryType, int], None]
        ] = None,
        builder: typing.Optional[Builder] = None,
        timeout: int = 1,
        max_iterations: typing.Optional[int] = 5,
        select_hits: typing.Optional[typing.Callable[["TopHits[_JACKHMMERQueryType]"], None]] = None,
        checkpoints: bool = False,
        **options,  # type: Unpack[PipelineOptions]
    ) -> None:
        super().__init__(
            queries=queries,
            targets=targets,
            cpus=cpus,
            callback=callback,
            builder=builder,
            timeout=timeout,
            **options,
        )
        self.max_iterations = max_iterations
        self.select_hits = select_hits
        self.checkpoints = checkpoints

    def _new_worker(
        self,
        query_queue: "queue.Queue[typing.Optional[_BaseChore[_JACKHMMERQueryType, _I]]]",
        query_count: "multiprocessing.Value[int]",  # type: ignore
        kill_switch: threading.Event,
    ) -> _JACKHMMERWorker[_I]:
        params = [
            self.targets,
            query_queue,
            query_count,
            kill_switch,
            self.callback,
            self.options,
            copy.copy(self.builder),
            self.max_iterations,
            self.select_hits,
            self.checkpoints,
        ]
        if self.backend == "threading":
            return _JACKHMMERThread(*params)
        elif self.backend == "multiprocessing":
            return _JACKHMMERProcess(*params)
        else:
            raise ValueError(f"Invalid backend for `jackhmmer`: {self.backend!r}")


# --- jackhmmer -----------------------------------------------------------------

@typing.overload
def jackhmmer(
    queries: typing.Union[_JACKHMMERQueryType, typing.Iterable[_JACKHMMERQueryType]],
    sequences: typing.Iterable[DigitalSequence],
    *,
    max_iterations: typing.Optional[int] = 5,
    select_hits: typing.Optional[typing.Callable[["TopHits[_JACKHMMERQueryType]"], None]] = None,
    checkpoints: "Literal[True]",
    cpus: int = 0,
    backend: "BACKEND" = "threading",
    callback: typing.Optional[typing.Callable[[_JACKHMMERQueryType, int], None]] = None,
    builder: typing.Optional[Builder] = None,
    **options,  # type: Unpack[PipelineOptions]
) -> typing.Iterator[typing.Iterable[IterationResult]]:
    ...


@typing.overload
def jackhmmer(
    queries: typing.Union[_JACKHMMERQueryType, typing.Iterable[_JACKHMMERQueryType]],
    sequences: typing.Iterable[DigitalSequence],
    *,
    max_iterations: typing.Optional[int] = 5,
    select_hits: typing.Optional[typing.Callable[["TopHits[_JACKHMMERQueryType]"], None]] = None,
    checkpoints: "Literal[False]",
    cpus: int = 0,
    backend: "BACKEND" = "threading",
    callback: typing.Optional[typing.Callable[[_JACKHMMERQueryType, int], None]] = None,
    builder: typing.Optional[Builder] = None,
    **options,  # type: Unpack[PipelineOptions]
) -> typing.Iterator[IterationResult]:
    ...


@typing.overload
def jackhmmer(
    queries: typing.Union[_JACKHMMERQueryType, typing.Iterable[_JACKHMMERQueryType]],
    sequences: typing.Union[typing.Iterable[DigitalSequence], "SequenceFile[DigitalSequence]"],
    *,
    max_iterations: typing.Optional[int] = 5,
    select_hits: typing.Optional[typing.Callable[["TopHits[_JACKHMMERQueryType]"], None]] = None,
    checkpoints: bool = False,
    cpus: int = 0,
    backend: "BACKEND" = "threading",
    callback: typing.Optional[typing.Callable[[_JACKHMMERQueryType, int], None]] = None,
    builder: typing.Optional[Builder] = None,
    **options,  # type: Unpack[PipelineOptions]
) -> typing.Union[
    typing.Iterator[IterationResult], typing.Iterator[typing.Iterable[IterationResult]]
]:
    ...


def jackhmmer(
    queries: typing.Union[_JACKHMMERQueryType, typing.Iterable[_JACKHMMERQueryType]],
    sequences: typing.Union[typing.Iterable[DigitalSequence], "SequenceFile[DigitalSequence]"],
    *,
    max_iterations: typing.Optional[int] = 5,
    select_hits: typing.Optional[typing.Callable[["TopHits[_JACKHMMERQueryType]"], None]] = None,
    checkpoints: bool = False,
    cpus: int = 0,
    backend: "BACKEND" = "threading",
    callback: typing.Optional[typing.Callable[[_JACKHMMERQueryType, int], None]] = None,
    builder: typing.Optional[Builder] = None,
    **options,  # type: Unpack[PipelineOptions]
) -> typing.Union[
    typing.Iterator[IterationResult], typing.Iterator[typing.Iterable[IterationResult]]
]:
    """Search protein sequences against a sequence database.

    Arguments:
        queries (iterable of `DigitalSequence`): The query sequences to search
            for in the sequence database. Passing a single sequence object
            is supported.
        sequences (iterable of `~pyhmmer.easel.DigitalSequence`): A database
            of sequences to query. If you plan on using the same sequences
            several times, consider storing them into a
            `~pyhmmer.easel.DigitalSequenceBlock` directly. `jackhmmer` also
            supports passing a `~pyhmmer.easel.SequenceFile` as the target
            sequences which will be open again at each iteration.
        max_iterations (`int`): The maximum number of iterations for the
            search. Hits will be returned early if the searched converged.
        select_hits (callable, optional): A function or callable object
            for manually selecting hits during each iteration. It should
            take a single `~pyhmmer.plan7.TopHits` argument and change the
            inclusion of individual hits with the `~pyhmmer.plan7.Hit.include`
            and `~pyhmmer.plan7.Hit.drop` methods of `~pyhmmer.plan7.Hit`
            objects.
        checkpoints (`bool`): A logical flag to return the results at each
            iteration 'checkpoint'. If `True`, then an iterable of up to
            ``max_iterations`` `~pyhmmer.plan7.IterationResult` will be
            returned, rather than just the final iteration. This is similar
            to ``--chkhmm`` amd ``--chkali`` flags from HMMER3's
            ``jackhmmer`` interface.
        cpus (`int`): The number of threads to run in parallel. Pass ``1`` to
            run everything in the main thread, ``0`` to automatically
            select a suitable number (using `psutil.cpu_count`), or any
            positive number otherwise.
        callback (callable): A callback that is called everytime a query is
            processed with two arguments: the query, and the total number
            of queries. This can be used to display progress in UI.
        builder (`~pyhmmer.plan7.Builder`, optional): A builder to configure
            how the queries are converted to HMMs. Passing `None` will create
            a default instance.
        backend (`str`): The parallel backend to use for workers to be
            executed. Supports ``threading`` to use thread-based parallelism,
            or ``multiprocessing`` to use process-based parallelism.

    Yields:
        `~pyhmmer.plan7.IterationResult`: An *iteration result* instance for
        each query, in the same order the queries were passed in the input.
        If ``checkpoint`` option is `True`, all iterations will be returned
        instead of the last one.

    Raises:
        `~pyhmmer.errors.AlphabetMismatch`: When any of the query sequence
            the profile or the optional builder do not share the same
            alphabet.

    Note:
        Any additional keyword arguments passed to the `jackhmmer` function
        will be passed transparently to the `~pyhmmer.plan7.Pipeline` to
        be created in each worker thread.

    Caution:
        Default values used for ``jackhmmer`` do not correspond to the
        default parameters used for creating a pipeline in the other cases.
        If no parameter value is given as a keyword argument, `jackhmmer`
        will create the pipeline with ``incE=0.001`` and ``incdomE=0.001``,
        where a default `~pyhmmer.plan7.Pipeline` would use ``incE=0.01``
        and ``incdomE=0.01``.

    .. versionadded:: 0.8.0

    """
    cpus = cpus if cpus > 0 else psutil.cpu_count(logical=False) or os.cpu_count() or 1
    options.setdefault("incE", 0.001)
    options.setdefault("incdomE", 0.001)
    alphabet = options.get("alphabet")

    if not isinstance(queries, collections.abc.Iterable):
        queries = (queries,)

    if isinstance(sequences, SequenceFile):
        if sequences.name is None:
            raise ValueError("expected named `SequenceFile` for targets")
        if not sequences.digital:
            raise ValueError("expected digital mode `SequenceFile` for targets")
        assert sequences.alphabet is not None
        alphabet = alphabet or sequences.alphabet
        targets = typing.cast(DigitalSequenceBlock, sequences.read_block())
    elif isinstance(sequences, DigitalSequenceBlock):
        alphabet = alphabet or sequences.alphabet
        targets = sequences
    else:
        sequences = peekable(sequences)
        try:
            alphabet = alphabet or sequences.peek().alphabet or Alphabet.amino()
            targets = DigitalSequenceBlock(alphabet, sequences)
        except StopIteration:
            alphabet = alphabet or Alphabet.amino()
            targets = DigitalSequenceBlock(alphabet)

    if builder is None:
        builder = Builder(
            alphabet,
            seed=options.get("seed", 42),
            architecture="hand",
        )

    if "alphabet" not in options:
        options["alphabet"] = alphabet
    dispatcher = _JACKHMMERDispatcher(  # type: ignore
        queries=queries,
        targets=targets,
        cpus=cpus,
        backend="threading",
        callback=callback,
        builder=builder,
        max_iterations=max_iterations,
        select_hits=select_hits,
        checkpoints=checkpoints,
        **options,
    )
    return dispatcher.run()




from libc.stdint cimport int32_t, int64_t, uint32_t, uint64_t
from libc.stdio cimport FILE

from libeasel cimport eslERRBUFSIZE
from libeasel.sqio cimport esl_sqio_s


cdef extern from "esl_sqio.h" nogil:

    const size_t MAX_DB_VOLUMES
    const size_t MAX_RESIDUE_COUNT

    ctypedef esl_sqncbi_vol_s ESL_SQNCBI_VOLUME
    cdef struct esl_sqncbi_vol_s:
        char* name
        uint32_t start_seq
        uint32_t end_seq
        uint32_t hdr_off
        uint32_t seq_off
        uint32_t amb_off


    ctypedef esl_sqncbi_s ESL_SQNCBI_DATA
    cdef struct esl_sqncbi_s:
        FILE      *fppin
        FILE      *fpphr
        FILE      *fppsq
        char[eslERRBUFSIZE] errbuf

        char *title
        int  version
        char *timestamp

        uint32_t   num_seq
        uint64_t   total_res
        uint32_t   max_seq

        uint32_t   hdr_off
        uint32_t   seq_off
        uint32_t   amb_off

        int        index
        uint32_t   vol_index
        uint32_t   roff
        uint32_t   hoff
        uint32_t   doff
        uint32_t   eoff

        uint32_t   index_start
        uint32_t   index_end
        uint32_t  *hdr_indexes
        uint32_t  *seq_indexes
        uint32_t  *amb_indexes

        uint32_t   volumes
        ESL_SQNCBI_VOLUME[MAX_DB_VOLUMES] vols

        unsigned char *hdr_buf
        unsigned char *hdr_ptr
        int            hdr_alloced

        char          *name_ptr
        int32_t        name_size
        char          *acc_ptr
        int32_t        acc_size
        int32_t        int_id
        char          *str_id_ptr
        int32_t        str_id_size

        uint32_t       seq_apos
        uint32_t       seq_alen
        uint32_t       seq_cpos
        int32_t        seq_L

        int            alphatype
        char          *alphasym


    int  esl_sqncbi_Open(char *seqfile, int format, esl_sqio_s *sqfp)



from libeasel cimport ESL_DSQ
from libhmmer.impl_neon.p7_oprofile cimport P7_OPROFILE


cdef extern from "impl_neon/impl_neon.h" nogil:

    cdef size_t p7O_EXTRA_SB = 17;

    cdef void impl_Init()
    cdef int p7_SSVFilter(const ESL_DSQ *dsq, int L, const P7_OPROFILE *om, float *ret_sc)



import abc
import pickle
import unittest

from pyhmmer.errors import AlphabetMismatch
from pyhmmer.easel import (
    Alphabet,
    GeneticCode,
    SequenceBlock,
    TextSequenceBlock,
    DigitalSequenceBlock,
    Sequence,
    TextSequence,
    DigitalSequence,
)


class _TestSequenceBlock(abc.ABC):

    @abc.abstractmethod
    def _new_sequence(self, name, seq):
        pass

    @abc.abstractmethod
    def _new_block(self, sequences=()):
        pass

    def test_truth(self):
        self.assertTrue(not self._new_block())
        seq1 = self._new_sequence(b"seq1", "ATGC")
        block = self._new_block([seq1])
        self.assertTrue(self._new_block([seq1]))

    def test_identity(self):
        self.assertIsNot(self._new_block(), self._new_block())

    def test_len(self):
        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")
        seq3 = self._new_sequence(b"seq3", "TTGA")

        self.assertEqual(len(self._new_block([])), 0)
        self.assertEqual(len(self._new_block([seq1])), 1)
        self.assertEqual(len(self._new_block([seq1, seq2, seq3])), 3)

    def test_append(self):
        block = self._new_block()
        self.assertEqual(len(block), 0)
        seq1 = self._new_sequence(b"seq1", "ATGC")
        block.append(seq1)
        self.assertEqual(len(block), 1)
        self.assertEqual(block[0], seq1)
        seq2 = self._new_sequence(b"seq2", "ATGCA")
        block.append(seq2)
        self.assertEqual(len(block), 2)
        self.assertEqual(block[1], seq2)

    def test_iter(self):
        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")
        seq3 = self._new_sequence(b"seq3", "TTGA")

        block = self._new_block([seq1, seq2, seq3])
        self.assertEqual(len(block), 3)

        iterator = iter(block)
        self.assertIs(next(iterator), seq1)
        self.assertIs(next(iterator), seq2)
        self.assertIs(next(iterator), seq3)
        self.assertRaises(StopIteration, next, iterator)

    def test_clear(self):
        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")

        block = self._new_block([seq1, seq2])
        self.assertEqual(len(block), 2)
        block.clear()
        self.assertEqual(len(block), 0)
        block.clear()
        self.assertEqual(len(block), 0)

        block = self._new_block()
        self.assertEqual(len(block), 0)
        block.clear()
        self.assertEqual(len(block), 0)

    def test_getitem(self):
        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")

        block = self._new_block([seq1, seq2])
        self.assertEqual(len(block), 2)

        self.assertIs(block[0], seq1)
        self.assertIs(block[1], seq2)
        self.assertIs(block[-1], seq2)
        self.assertIs(block[-2], seq1)

        with self.assertRaises(IndexError):
            block[3]
        with self.assertRaises(IndexError):
            block[-5]

    def test_setitem(self):
        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")
        seq3 = self._new_sequence(b"seq3", "TTGA")

        block = self._new_block([seq1, seq2])
        self.assertEqual(len(block), 2)

        block[0] = seq3
        self.assertIs(block[0], seq3)
        self.assertIs(block[1], seq2)

    def test_remove(self):
        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")

        block = self._new_block([seq1, seq2])
        block.remove(seq1)
        self.assertEqual(len(block), 1)
        self.assertIs(block[0], seq2)

    def test_index(self):
        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")
        seq3 = self._new_sequence(b"seq3", "TTGA")

        block = self._new_block([seq1, seq2])
        self.assertEqual(len(block), 2)

        self.assertEqual(block.index(seq1), 0)
        self.assertEqual(block.index(seq2), 1)

        self.assertRaises(ValueError, block.index, seq3)
        self.assertRaises(ValueError, block.index, seq1, start=1)

    def test_pop(self):
        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")
        seq3 = self._new_sequence(b"seq3", "TTGA")

        block = self._new_block([seq1, seq2, seq3])
        self.assertEqual(len(block), 3)

        self.assertIs(block.pop(), seq3)
        self.assertIs(block.pop(0), seq1)

        self.assertEqual(len(block), 1)
        self.assertIs(block[0], seq2)

        self.assertIs(block.pop(-1), seq2)
        self.assertEqual(len(block), 0)
        self.assertRaises(IndexError, block.pop)

        empty = self._new_block()
        self.assertRaises(IndexError, block.pop)

    def test_setitem_slice(self):
        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")
        seq3 = self._new_sequence(b"seq3", "TTGA")

        block = self._new_block([seq1, seq2])
        self.assertEqual(len(block), 2)
        self.assertIs(block[0], seq1)
        self.assertIs(block[1], seq2)

        block[:] = [seq1, seq2, seq3]
        self.assertEqual(len(block), 3)
        self.assertIs(block[0], seq1)
        self.assertIs(block[1], seq2)
        self.assertIs(block[2], seq3)

        block[2:5] = [seq3, seq3, seq3]
        self.assertEqual(len(block), 5)
        self.assertIs(block[3], seq3)
        self.assertIs(block[4], seq3)

    def test_delitem_slice(self):
        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")
        seq3 = self._new_sequence(b"seq3", "TTGA")

        block = self._new_block([seq1, seq2, seq3])
        self.assertEqual(len(block), 3)
        del block[:]
        self.assertEqual(len(block), 0)

        block = self._new_block([seq1, seq2, seq3])
        self.assertEqual(len(block), 3)
        del block[::2]
        self.assertEqual(len(block), 1)
        self.assertIs(block[0], seq2)

    def test_contains(self):
        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")
        seq3 = self._new_sequence(b"seq3", "TTGA")

        block = self._new_block([seq1, seq2])
        self.assertEqual(len(block), 2)
        self.assertIn(seq1, block)
        self.assertIn(seq2, block)
        self.assertNotIn(seq3, block)

        self.assertNotIn(42, block)
        self.assertNotIn(object(), block)

    def test_largest(self):
        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")
        seq3 = self._new_sequence(b"seq3", "TTGA")

        block = self._new_block([seq1, seq2])
        self.assertIs(block.largest(), seq2)

        block.pop(1)
        self.assertIs(block.largest(), seq1)

        with self.assertRaises(ValueError):
            block = self._new_block([])
            block.largest()

    def test_copy(self):
        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")
        seq3 = self._new_sequence(b"seq3", "TTGA")

        block = self._new_block([seq1, seq2, seq3])
        block_copy = block.copy()

        self.assertEqual(len(block), len(block_copy))
        self.assertEqual(block, block_copy)

        self.assertIs(block[0], block_copy[0])
        self.assertIs(block[1], block_copy[1])
        self.assertIs(block[2], block_copy[2])

        block.remove(seq3)
        self.assertEqual(len(block), 2)
        self.assertEqual(len(block_copy), 3)

    def test_pickle(self):
        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")
        seq3 = self._new_sequence(b"seq3", "TTGA")

        block = self._new_block([seq1, seq2, seq3])
        block_pickled = pickle.loads(pickle.dumps(block))
        
        self.assertEqual(len(block), len(block_pickled))
        self.assertEqual(block[0], block_pickled[0])
        self.assertEqual(block[1], block_pickled[1])
        self.assertEqual(block[2], block_pickled[2])

    def test_indexed(self):
        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")
        seq3 = self._new_sequence(b"seq3", "TTGA")

        block = self._new_block([seq1, seq2, seq3])

        self.assertEqual(len(block.indexed), 3)
        self.assertEqual(block.indexed[b"seq1"], block[0])
        self.assertEqual(block.indexed[b"seq2"], block[1])
        self.assertEqual(block.indexed[b"seq3"], block[2])

    def test_indexed_duplicates(self):
        seq1a = self._new_sequence(b"seq1", "ATGC")
        seq1b = self._new_sequence(b"seq1", "ATGCA")

        block = self._new_block([seq1a, seq1b])

        with self.assertRaises(KeyError):
            block.indexed[b"seq1"]

    def test_indexed_key_error(self):
        block = self._new_block([])
        with self.assertRaises(KeyError):
            block.indexed[b"xxx"]

    def test_indexed_type_error(self):
        block = self._new_block([])
        with self.assertRaises(TypeError):
            block.indexed[1]    


class TestTextSequenceBlock(_TestSequenceBlock, unittest.TestCase):

    def _new_sequence(self, name, seq):
        return TextSequence(name=name, sequence=seq)

    def _new_block(self, sequences=()):
        return TextSequenceBlock(sequences)

    def test_digitize(self):
        alphabet = Alphabet.dna()

        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")

        block = self._new_block([seq1, seq2])
        self.assertEqual(len(block), 2)

        dblock = block.digitize(alphabet)
        self.assertEqual(len(dblock), 2)
        self.assertEqual(dblock[0], seq1.digitize(alphabet))
        self.assertEqual(dblock[1], seq2.digitize(alphabet))


class TestDigitalSequenceBlock(_TestSequenceBlock, unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.alphabet = Alphabet.dna()

    def _new_sequence(
        self,
        name,
        seq,
        description=None,
        source=None,
    ):
        return TextSequence(
            name=name,
            sequence=seq,
            description=description,
            source=source,
        ).digitize(self.alphabet)

    def _new_block(self, sequences=()):
        return DigitalSequenceBlock(self.alphabet, sequences)

    def test_textize(self):
        seq1 = self._new_sequence(b"seq1", "ATGC")
        seq2 = self._new_sequence(b"seq2", "ATGCA")

        block = self._new_block([seq1, seq2])
        self.assertEqual(len(block), 2)

        tblock = block.textize()
        self.assertEqual(len(tblock), 2)
        self.assertEqual(tblock[0], seq1.textize())
        self.assertEqual(tblock[1], seq2.textize())

    def test_translate(self):
        seq1 = self._new_sequence(
            b"seq1",
            "ATGCTG",
            description=b"one",
            source=b"some_source"
        )
        seq2 = self._new_sequence(
            b"seq2",
            "ATGCCC",
            description=b"two",
            source=b"other_source"
        )
        seq3 = self._new_sequence(
            b"seq3",
            "ATTCTGATGATA",
        )

        block = self._new_block([seq1, seq2, seq3])
        prots = block.translate().textize()

        self.assertEqual(len(prots), 3)
        # test sequences
        self.assertEqual(prots[0].sequence, "ML")
        self.assertEqual(prots[1].sequence, "MP")
        self.assertEqual(prots[2].sequence, "ILMI")
        # test names
        self.assertEqual(prots[0].name, b"seq1")
        self.assertEqual(prots[1].name, b"seq2")
        self.assertEqual(prots[2].name, b"seq3")
        # test other metadata fields
        self.assertEqual(prots[0].description, b"one")
        self.assertEqual(prots[1].description, b"two")
        self.assertEqual(prots[2].description, b"")
        self.assertEqual(prots[0].source, b"some_source")
        self.assertEqual(prots[1].source, b"other_source")
        self.assertEqual(prots[2].source, b"")

    def test_translate_empty(self):
        gencode = GeneticCode()
        block = DigitalSequenceBlock(gencode.nucleotide_alphabet)
        prots = block.translate(gencode)
        self.assertEqual(len(prots), 0)

    def test_translate_alphabet_mismatch(self):
        gencode = GeneticCode()
        block = DigitalSequenceBlock(gencode.amino_alphabet)
        with self.assertRaises(AlphabetMismatch):
            prots = block.translate(gencode)


from libc.stdint cimport int64_t, uint8_t, uint32_t, uint64_t
from libc.stdio cimport FILE

from libeasel cimport esl_pos_t
from libeasel.rand64 cimport ESL_RAND64
from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.getopts cimport ESL_GETOPTS
from libeasel.keyhash cimport ESL_KEYHASH
from libeasel.msa cimport ESL_MSA
from libeasel.sq cimport ESL_SQ
from libhmmer.p7_domain cimport P7_DOMAIN
from libhmmer.p7_pipeline cimport p7_pipemodes_e, P7_PIPELINE
from libhmmer.p7_trace cimport P7_TRACE
from libhmmer.p7_alidisplay cimport P7_ALIDISPLAY


cdef extern from "hmmer.h" nogil:

    cdef enum p7_hitflags_e:
        p7_HITFLAGS_DEFAULT = 0b00000
        p7_IS_INCLUDED      = 0b00001
        p7_IS_REPORTED      = 0b00010
        p7_IS_NEW           = 0b00100
        p7_IS_DROPPED       = 0b01000
        p7_IS_DUPLICATE     = 0b10000

    ctypedef struct P7_HIT:
        char* name
        char* acc
        char* desc
        int window_length
        double sortkey

        float score
        float pre_score
        float sum_score

        double lnP
        double pre_lnP
        double sum_lnP

        float nexpected
        int nregions
        int nclustered
        int noverlaps
        int nenvelopes
        int ndom

        uint32_t flags
        int nreported
        int nincluded
        int best_domain

        int64_t seqidx
        int64_t subseq_start

        P7_DOMAIN *dcl
        esl_pos_t offset

    cdef P7_HIT *p7_hit_Create_empty()
    cdef void p7_hit_Destroy(P7_HIT *the_hit)
    cdef int p7_hit_Serialize(const P7_HIT *obj, uint8_t **buf, uint32_t *n, uint32_t *nalloc)
    cdef int p7_hit_Deserialize(const uint8_t *buf, uint32_t *n, P7_HIT *ret_obj)
    cdef int p7_hit_TestSample(ESL_RAND64 *rng, P7_HIT **ret_obj)
    cdef int p7_hit_Compare(P7_HIT *first, P7_HIT *second, double atol, double rtol)



import os

try:
    from importlib.resources import files as resource_files
except ImportError:
    try:
        from importlib_resources import files as resource_files
    except ImportError:
        resource_files = None

EASEL_FOLDER = os.path.realpath(
    os.path.join(
        __file__,
        os.pardir,
        os.pardir,
        os.pardir,
        os.pardir,
        "vendor",
        "easel",
    )
)

HMMER_FOLDER = os.path.realpath(
    os.path.join(
        __file__,
        os.pardir,
        os.pardir,
        os.pardir,
        os.pardir,
        "vendor",
        "hmmer",
    )
)



import copy
import io
import itertools
import os
import shutil
import unittest
import tempfile

import pyhmmer
from pyhmmer.errors import EaselError, AlphabetMismatch
from pyhmmer.easel import Alphabet, SequenceFile, Randomness
from pyhmmer.plan7 import HMM, HMMFile, Profile, Background


class TestOptimizedProfile(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.rng = Randomness(seed=0)
        cls.alphabet = Alphabet.amino()
        cls.background = Background(cls.alphabet)

    @classmethod
    def _random_hmm(cls, name, M=100):
        hmm = HMM.sample(cls.alphabet, M, cls.rng)
        hmm.name = name
        return hmm

    @classmethod
    def _random_profile(cls, name, M=100):
        hmm = cls._random_hmm(name, M=M)
        return hmm.to_profile(L=20)

    @classmethod
    def _random_optimized_profile(cls, name, M=100):
        return cls._random_profile(name, M=M).to_optimized()

    def test_eq(self):
        profile1 = self._random_profile(b"profile1")
        om1 = profile1.to_optimized()
        om2 = profile1.to_optimized()
        self.assertEqual(om1, om2)
        profile2 = self._random_profile(b"profile2")
        om3 = profile2.to_optimized()
        self.assertNotEqual(om1, om3)

    def test_copy(self):
        om1 = self._random_optimized_profile(b"profile1")
        om2 = copy.copy(om1)
        self.assertEqual(om1, om2)
        self.assertIsNot(om1, om2)

    def test_M(self):
        om = self._random_optimized_profile(b"profile1", 20)
        self.assertEqual(om.M, 20)

    def test_L(self):
        hmm = self._random_hmm(b"profile1")
        profile = hmm.to_profile(self.background, L=40)
        om = profile.to_optimized()
        self.assertEqual(om.L, 40)
        om.L = 30
        self.assertEqual(om.L, 30)

    def test_accession(self):
        hmm = self._random_hmm(b"profile1")
        hmm.accession = b"TST001"
        profile = hmm.to_profile(self.background, L=10)
        om = profile.to_optimized()
        self.assertEqual(om.accession, hmm.accession)

    def test_description(self):
        hmm = self._random_hmm(b"profile1")
        hmm.description = b"test profile one"
        profile = hmm.to_profile(self.background, L=10)
        om = profile.to_optimized()
        self.assertEqual(om.description, hmm.description)

    def test_name(self):
        hmm = self._random_hmm(b"profile1")
        profile = hmm.to_profile(self.background, L=10)
        om = profile.to_optimized()
        self.assertEqual(om.name, hmm.name)

    def test_consensus(self):
        hmm = self._random_hmm(b"profile1")
        profile = hmm.to_profile(self.background, L=10)
        om1 = profile.to_optimized()
        self.assertEqual(om1.consensus, hmm.consensus)
        profile = Profile(hmm.M, hmm.alphabet)
        om2 = profile.to_optimized()
        self.assertIs(om2.consensus, None)

    def test_consensus_structure(self):
        profile = Profile(100, self.alphabet)
        om = profile.to_optimized()
        self.assertIs(om.consensus_structure, None)

    def test_offsets(self):
        om = self._random_optimized_profile(b"profile1")
        self.assertIs(om.offsets.model, None)
        self.assertIs(om.offsets.profile, None)
        self.assertIs(om.offsets.filter, None)



from . import (
    test_doctest,
    test_easel,
    test_errors,
    test_hmmer,
    test_plan7,
)

def load_tests(loader, suite, pattern):
    suite.addTests(loader.loadTestsFromModule(test_doctest))
    suite.addTests(loader.loadTestsFromModule(test_easel))
    suite.addTests(loader.loadTestsFromModule(test_errors))
    suite.addTests(loader.loadTestsFromModule(test_hmmer))
    suite.addTests(loader.loadTestsFromModule(test_plan7))
    return suite



import copy
import io
import itertools
import os
import shutil
import unittest
import tempfile

import pyhmmer
from pyhmmer.errors import EaselError, AlphabetMismatch
from pyhmmer.easel import Alphabet, SequenceFile, TextSequence, DigitalMSA, TextMSA
from pyhmmer.plan7 import HMM, HMMFile, TraceAligner, Traces, Trace

from .. import __name__ as __package__
from .utils import HMMER_FOLDER, resource_files


class TestTraces(unittest.TestCase):

    def _new_trace(self, n=3):
        seq = TextSequence(sequence="N"*n)
        trace = Trace.from_sequence(seq)
        return trace

    def _new_block(self, traces=()):
        return Traces(traces)

    @unittest.skipUnless(resource_files, "importlib.resources.files not available")
    def test_aligner_traces(self):
        hmm_path = resource_files(__package__).joinpath("data", "hmms", "txt", "LuxC.hmm")
        seqs_path = resource_files(__package__).joinpath("data", "seqs", "LuxC.faa")

        with HMMFile(hmm_path) as hmm_file:
            hmm = next(hmm_file)
        with SequenceFile(seqs_path, digital=True, alphabet=hmm.alphabet) as seqs_file:
            seqs = seqs_file.read_block()

        aligner = TraceAligner()
        traces = aligner.compute_traces(hmm, seqs)

        for trace, seq in zip(traces, seqs):
            self.assertEqual(trace.M, hmm.M)
            self.assertEqual(trace.L, len(seq))

    def test_bool(self):
        self.assertTrue(not self._new_block())
        self.assertFalse(self._new_block())
        t1 = self._new_trace()
        block = self._new_block([t1])
        self.assertTrue(block)

    def test_identity(self):
        self.assertIsNot(self._new_block(), self._new_block())

    def test_len(self):
        t1 = self._new_trace(1)
        t2 = self._new_trace(2)
        t3 = self._new_trace(3)

        self.assertEqual(len(self._new_block([])), 0)
        self.assertEqual(len(self._new_block([t1])), 1)
        self.assertEqual(len(self._new_block([t1, t2, t3])), 3)

    def test_append(self):
        block = self._new_block()
        self.assertEqual(len(block), 0)
        t1 = self._new_trace(1)
        block.append(t1)
        self.assertEqual(len(block), 1)
        self.assertEqual(block[0], t1)
        t2 = self._new_trace(2)
        block.append(t2)
        self.assertEqual(len(block), 2)
        self.assertEqual(block[1], t2)

    def test_iter(self):
        t1 = self._new_trace(1)
        t2 = self._new_trace(2)
        t3 = self._new_trace(3)

        block = self._new_block([t1, t2, t3])
        self.assertEqual(len(block), 3)

        iterator = iter(block)
        self.assertIs(next(iterator), t1)
        self.assertIs(next(iterator), t2)
        self.assertIs(next(iterator), t3)
        self.assertRaises(StopIteration, next, iterator)

    def test_clear(self):
        t1 = self._new_trace(1)
        t2 = self._new_trace(2)

        block = self._new_block([t1, t2])
        self.assertEqual(len(block), 2)
        block.clear()
        self.assertEqual(len(block), 0)
        block.clear()
        self.assertEqual(len(block), 0)

        block = self._new_block()
        self.assertEqual(len(block), 0)
        block.clear()
        self.assertEqual(len(block), 0)

    def test_getitem(self):
        t1 = self._new_trace(1)
        t2 = self._new_trace(2)

        block = self._new_block([t1, t2])
        self.assertEqual(len(block), 2)

        self.assertIs(block[0], t1)
        self.assertIs(block[1], t2)
        self.assertIs(block[-1], t2)
        self.assertIs(block[-2], t1)

    def test_getitem_indexerror(self):
        t1 = self._new_trace()
        t2 = self._new_trace()
        block = self._new_block([t1, t2])

        with self.assertRaises(IndexError):
            block[3]
        with self.assertRaises(IndexError):
            block[-5]

    def test_setitem(self):
        t1 = self._new_trace(1)
        t2 = self._new_trace(2)
        t3 = self._new_trace(3)

        block = self._new_block([t1, t2])
        self.assertEqual(len(block), 2)

        block[0] = t3
        self.assertIs(block[0], t3)
        self.assertIs(block[1], t2)

    def test_remove(self):
        t1 = self._new_trace(1)
        t2 = self._new_trace(2)

        block = self._new_block([t1, t2])
        block.remove(t1)
        self.assertEqual(len(block), 1)
        self.assertIs(block[0], t2)

    def test_index(self):
        t1 = self._new_trace(1)
        t2 = self._new_trace(2)
        t3 = self._new_trace(3)

        block = self._new_block([t1, t2])
        self.assertEqual(len(block), 2)

        self.assertEqual(block.index(t1), 0)
        self.assertEqual(block.index(t2), 1)

        self.assertRaises(ValueError, block.index, t3)
        self.assertRaises(ValueError, block.index, t1, start=1)

    def test_pop(self):
        t1 = self._new_trace(1)
        t2 = self._new_trace(2)
        t3 = self._new_trace(3)

        block = self._new_block([t1, t2, t3])
        self.assertEqual(len(block), 3)

        self.assertIs(block.pop(), t3)
        self.assertIs(block.pop(0), t1)

        self.assertEqual(len(block), 1)
        self.assertIs(block[0], t2)

        self.assertIs(block.pop(-1), t2)
        self.assertEqual(len(block), 0)
        self.assertRaises(IndexError, block.pop)

        empty = self._new_block()
        self.assertRaises(IndexError, block.pop)

    def test_setitem_slice(self):
        t1 = self._new_trace(1)
        t2 = self._new_trace(2)
        t3 = self._new_trace(3)

        block = self._new_block([t1, t2])
        self.assertEqual(len(block), 2)
        self.assertIs(block[0], t1)
        self.assertIs(block[1], t2)

        block[:] = [t1, t2, t3]
        self.assertEqual(len(block), 3)
        self.assertIs(block[0], t1)
        self.assertIs(block[1], t2)
        self.assertIs(block[2], t3)

        block[2:5] = [t3, t3, t3]
        self.assertEqual(len(block), 5)
        self.assertIs(block[3], t3)
        self.assertIs(block[4], t3)

    # def test_delitem_slice(self):
    #     t1 = self._new_trace(1)
    #     t2 = self._new_trace(2)
    #     t3 = self._new_trace(3)

    #     block = self._new_block([t1, t2, t3])
    #     self.assertEqual(len(block), 3)
    #     del block[:]
    #     self.assertEqual(len(block), 0)

    #     block = self._new_block([t1, t2, t3])
    #     self.assertEqual(len(block), 3)
    #     del block[::2]
    #     self.assertEqual(len(block), 1)
    #     self.assertIs(block[0], t2)

    def test_contains(self):
        t1 = self._new_trace(1)
        t2 = self._new_trace(2)
        t3 = self._new_trace(3)

        block = self._new_block([t1, t2])
        self.assertEqual(len(block), 2)
        self.assertIn(t1, block)
        self.assertIn(t2, block)
        self.assertNotIn(t3, block)

        self.assertNotIn(42, block)
        self.assertNotIn(object(), block)

    # def test_copy(self):
    #     t1 = self._new_trace()
    #     t2 = self._new_trace()
    #     t3 = self._new_trace()

    #     block = self._new_block([t1, t2, t3])
    #     block_copy = block.copy()

    #     self.assertEqual(len(block), len(block_copy))
    #     self.assertEqual(block, block_copy)

    #     self.assertIs(block[0], block_copy[0])
    #     self.assertIs(block[1], block_copy[1])
    #     self.assertIs(block[2], block_copy[2])

    #     block.remove(t3)
    #     self.assertEqual(len(block), 2)
    #     self.assertEqual(len(block_copy), 3)

    # def test_pickle(self):
    #     t1 = self._new_trace(1)
    #     t2 = self._new_trace(2)
    #     t3 = self._new_trace(3)

    #     block = self._new_block([t1, t2, t3])
    #     block_pickled = pickle.loads(pickle.dumps(block))
        
    #     self.assertEqual(len(block), len(block_pickled))
    #     self.assertEqual(block[0], block_pickled[0])
    #     self.assertEqual(block[1], block_pickled[1])
    #     self.assertEqual(block[2], block_pickled[2])



cdef extern from "esl_mixgev.h" nogil:

    ctypedef struct ESL_MIXGEV:
        double* q
        double* mu
        double* lambda_
        double* alpha
        double* wrk
        bint* isgumbel
        int K

    ESL_MIXGEV *esl_mixgev_Create(int K)
    void        esl_mixgev_Destroy(ESL_MIXGEV *mg)
    int         esl_mixgev_Copy(ESL_MIXGEV *dest, ESL_MIXGEV *src)
    int         esl_mixgev_ForceGumbel(ESL_MIXGEV *mg, int which)

    double      esl_mixgev_pdf    (double x, ESL_MIXGEV *mg)
    double      esl_mixgev_logpdf (double x, ESL_MIXGEV *mg)
    double      esl_mixgev_cdf    (double x, ESL_MIXGEV *mg)
    double      esl_mixgev_logcdf (double x, ESL_MIXGEV *mg)
    double      esl_mixgev_surv   (double x, ESL_MIXGEV *mg)
    double      esl_mixgev_logsurv(double x, ESL_MIXGEV *mg)
    double      esl_mixgev_invcdf (double p, ESL_MIXGEV *mg)

    double      esl_mixgev_generic_pdf   (double x, void *params)
    double      esl_mixgev_generic_cdf   (double x, void *params)
    double      esl_mixgev_generic_surv  (double x, void *params)
    double      esl_mixgev_generic_invcdf(double p, void *params)

    int         esl_mixgev_Plot(FILE *fp, ESL_MIXGEV *mg,
    				   double (*func)(double x, ESL_MIXGEV *mg),
    				   double xmin, double xmax, double xstep)

    double      esl_mixgev_Sample(ESL_RANDOMNESS *r, ESL_MIXGEV *mg)
    int         esl_mixgev_FitGuess(ESL_RANDOMNESS *r, double *x, int n
    				       ESL_MIXGEV *mg)

    int         esl_mixgev_FitComplete(double *x, int n, ESL_MIXGEV *mg)



import copy
import gc
import io
import os
import pickle
import tempfile
import unittest

from pyhmmer import easel, plan7


class TestBuilder(unittest.TestCase):

    def test_init(self):
        abc = easel.Alphabet.dna()
        bg = plan7.Background(abc)
        self.assertFalse(bg.uniform)

        bgu = plan7.Background(abc, uniform=True)
        self.assertTrue(bgu.uniform)

    def test_copy(self):
        abc = easel.Alphabet.dna()
        bg = plan7.Background(abc)

        bg_copy = copy.copy(bg)
        self.assertIs(bg.alphabet, bg_copy.alphabet)
        self.assertEqual(bg.uniform, bg_copy.uniform)
        self.assertEqual(bg.L, bg_copy.L)
        self.assertEqual(bg.transition_probability, bg_copy.transition_probability)
        self.assertEqual(bg.omega, bg_copy.omega)
        self.assertEqual(list(bg.residue_frequencies), list(bg_copy.residue_frequencies))

    def test_pickle(self):
        abc = easel.Alphabet.dna()
        bg = plan7.Background(abc)

        bg2 = pickle.loads(pickle.dumps(bg))
        self.assertEqual(bg.alphabet, bg2.alphabet)
        self.assertEqual(bg.transition_probability, bg2.transition_probability)
        self.assertEqual(bg.residue_frequencies, bg2.residue_frequencies)

    def test_repr(self):
        abc = easel.Alphabet.amino()
        bg = plan7.Background(abc)
        self.assertEqual(repr(bg), "Background(Alphabet.amino(), uniform=False)")
        bg2 = plan7.Background(abc, uniform=True)
        self.assertEqual(repr(bg2), "Background(Alphabet.amino(), uniform=True)")



from libc.stdio cimport FILE
from libc.stdint cimport uint32_t


cdef extern from "esl_random.h" nogil:

    cdef enum esl_randomness_type:
        eslRND_FAST = 0
        eslRND_MERSENNE = 1

    ctypedef struct ESL_RANDOMNESS:
        esl_randomness_type      type
        int      mti
        uint32_t[624] mt
        uint32_t x
        uint32_t seed

    # The ESL_RANDOMNESS object.
    ESL_RANDOMNESS *esl_randomness_Create    (uint32_t seed);
    ESL_RANDOMNESS *esl_randomness_CreateFast(uint32_t seed);   # DEPRECATED. Use esl_randomness_Create.  The Knuth LCG used to have a speed advantage for us, but MT is fast.
    ESL_RANDOMNESS *esl_randomness_CreateTimeseeded();          # DEPRECATED. Use esl_randomness_Create(0)
    void            esl_randomness_Destroy(ESL_RANDOMNESS *r);
    int             esl_randomness_Init(ESL_RANDOMNESS *r, uint32_t seed);
    uint32_t        esl_randomness_GetSeed(const ESL_RANDOMNESS *r);

    # The generator, esl_random().
    double   esl_random       (ESL_RANDOMNESS *r);
    uint32_t esl_random_uint32(ESL_RANDOMNESS *r);

    uint32_t esl_rnd_mix3(uint32_t a, uint32_t b, uint32_t c);

    # Debugging/development tools.
    int esl_randomness_Dump(FILE *fp, ESL_RANDOMNESS *r);

    # Other fundamental sampling (including Gaussian, gamma).
    double esl_rnd_UniformPositive(ESL_RANDOMNESS *r);
    double esl_rnd_Gaussian (ESL_RANDOMNESS *rng, double mean, double stddev);
    double esl_rnd_Gamma    (ESL_RANDOMNESS *rng, double a);
    int    esl_rnd_Dirichlet(ESL_RANDOMNESS *rng, const double *alpha, int K, double *p);  # Pass alpha=NULL if you just want a uniform draw.
    int    esl_rnd_Deal     (ESL_RANDOMNESS *rng, int m, int n, int *deal);

    # Multinomial sampling from discrete probability n-vectors.
    int    esl_rnd_DChoose   (ESL_RANDOMNESS *r, const double *p,   int N);
    int    esl_rnd_FChoose   (ESL_RANDOMNESS *r, const float  *p,   int N);
    int    esl_rnd_DChooseCDF(ESL_RANDOMNESS *r, const double *cdf, int N);
    int    esl_rnd_FChooseCDF(ESL_RANDOMNESS *r, const float  *cdf, int N);

    # Random data generators (unit testing, etc.)
    int    esl_rnd_mem        (ESL_RANDOMNESS *rng, void *buf, int n);
    int    esl_rnd_floatstring(ESL_RANDOMNESS *rng, char *s);



import collections
import copy
import os
import queue
import multiprocessing
import typing
import threading

import psutil

from ..easel import Alphabet, DigitalSequence, DigitalMSA, DigitalSequenceBlock, SequenceFile
from ..plan7 import TopHits, Builder, Pipeline
from ..utils import singledispatchmethod, peekable
from ._base import _BaseDispatcher, _BaseWorker, _BaseChore

_PHMMERQueryType = typing.Union[DigitalSequence, DigitalMSA]
_M = typing.TypeVar("_M", DigitalSequence, DigitalMSA)

if typing.TYPE_CHECKING:
    from ._base import Unpack, PipelineOptions, BACKEND

# --- Worker -------------------------------------------------------------------

class _PHMMERWorker(
    _BaseWorker[
        _PHMMERQueryType,
        typing.Union[DigitalSequenceBlock, "SequenceFile[DigitalSequence]"],
        "TopHits[_PHMMERQueryType]",
    ],
    threading.Thread
):
    @singledispatchmethod
    def query(self, query) -> "TopHits[Any]":  # type: ignore
        raise TypeError(
            "Unsupported query type for `phmmer`: {}".format(type(query).__name__)
        )

    @query.register(DigitalSequence)
    def _(self, query: DigitalSequence) -> "TopHits[DigitalSequence]":  # type: ignore
        assert self.pipeline is not None
        return self.pipeline.search_seq(query, self.targets, self.builder)

    @query.register(DigitalMSA)
    def _(self, query: DigitalMSA) -> "TopHits[DigitalMSA]":  # type: ignore
        assert self.pipeline is not None
        return self.pipeline.search_msa(query, self.targets, self.builder)


class _PHMMERThread(_PHMMERWorker, threading.Thread):
    pass


class _PHMMERProcess(_PHMMERWorker, multiprocessing.Process):
    pass

# --- Dispatcher ---------------------------------------------------------------

class _PHMMERDispatcher(
    _BaseDispatcher[
        _PHMMERQueryType,
        typing.Union[DigitalSequenceBlock, "SequenceFile[DigitalSequence]"],
        "TopHits[_PHMMERQueryType]",
    ]
):
    def _new_worker(
        self,
        query_queue: "queue.Queue[typing.Optional[_BaseChore[_PHMMERQueryType, TopHits[_PHMMERQueryType]]]]",
        query_count: "multiprocessing.Value[int]",  # type: ignore
        kill_switch: threading.Event,
    ) -> _PHMMERWorker:
        if isinstance(self.targets, SequenceFile):
            assert self.targets.name is not None
            targets = SequenceFile(
                self.targets.name,
                format=self.targets.format,
                digital=True,
                alphabet=self.options["alphabet"],
            )
        else:
            targets = self.targets  # type: ignore
        params = [
            targets,
            query_queue,
            query_count,
            kill_switch,
            self.callback,
            self.options,
            copy.copy(self.builder),
        ]
        if self.backend == "threading":
            return _PHMMERThread(*params)
        elif self.backend == "multiprocessing":
            return _PHMMERProcess(*params)
        else:
            raise ValueError(f"Invalid backend for `phmmer`: {self.backend!r}")

# --- phmmer -----------------------------------------------------------------

def phmmer(
    queries: typing.Union[_M, typing.Iterable[_M]],
    sequences: typing.Iterable[DigitalSequence],
    *,
    cpus: int = 0,
    callback: typing.Optional[typing.Callable[[_M, int], None]] = None,
    backend: "BACKEND" = "threading",
    builder: typing.Optional[Builder] = None,
    **options,  # type: Unpack[PipelineOptions]
) -> typing.Iterator["TopHits[_M]"]:
    """Search protein sequences against a sequence database.

    Arguments:
        queries (iterable of `DigitalSequence` or `DigitalMSA`): The query
            sequences to search for in the sequence database. Passing a
            single object is supported.
        sequences (iterable of `~pyhmmer.easel.DigitalSequence`): A database
            of sequences to query. If you plan on using the same sequences
            several times, consider storing them into a
            `~pyhmmer.easel.DigitalSequenceBlock` directly. If a
            `SequenceFile` is given, profiles will be loaded iteratively
            from disk rather than prefetched.
        cpus (`int`): The number of threads to run in parallel. Pass ``1`` to
            run everything in the main thread, ``0`` to automatically
            select a suitable number (using `psutil.cpu_count`), or any
            positive number otherwise.
        callback (callable): A callback that is called everytime a query is
            processed with two arguments: the query, and the total number
            of queries. This can be used to display progress in UI.
        builder (`~pyhmmer.plan7.Builder`, optional): A builder to configure
            how the queries are converted to HMMs. Passing `None` will create
            a default instance.
        backend (`str`): The parallel backend to use for workers to be
            executed. Supports ``threading`` to use thread-based parallelism,
            or ``multiprocessing`` to use process-based parallelism.

    Yields:
        `~pyhmmer.plan7.TopHits`: A *top hits* instance for each query,
        in the same order the queries were passed in the input.

    Raises:
        `~pyhmmer.errors.AlphabetMismatch`: When any of the query sequence
            the profile or the optional builder do not share the same
            alphabet.

    Note:
        Any additional keyword arguments passed to the `phmmer` function
        will be passed transparently to the `~pyhmmer.plan7.Pipeline` to
        be created in each worker thread.

    .. versionadded:: 0.2.0

    .. versionchanged:: 0.3.0
       Allow using `DigitalMSA` queries.

    .. versionchanged:: 0.7.0
        Queries may now be an iterable of different types, or a single object.

    """
    cpus = cpus if cpus > 0 else psutil.cpu_count(logical=False) or os.cpu_count() or 1
    alphabet = options.get("alphabet")

    if not isinstance(queries, collections.abc.Iterable):
        queries = (queries,)

    if isinstance(sequences, SequenceFile):
        if sequences.name is None:
            raise ValueError("expected named `SequenceFile` for targets")
        if not sequences.digital:
            raise ValueError("expected digital mode `SequenceFile` for targets")
        assert sequences.alphabet is not None
        alphabet = alphabet or sequences.alphabet
        targets: typing.Union["SequenceFile[DigitalSequence]", DigitalSequenceBlock] = sequences
    elif isinstance(sequences, DigitalSequenceBlock):
        alphabet = alphabet or sequences.alphabet
        targets = sequences
    else:
        sequences = peekable(sequences)
        try:
            alphabet = alphabet or sequences.peek().alphabet or Alphabet.amino()
            targets = DigitalSequenceBlock(alphabet, sequences)
        except StopIteration:
            alphabet = alphabet or Alphabet.amino()
            targets = DigitalSequenceBlock(alphabet)

    if "alphabet" not in options:
        options["alphabet"] = alphabet
    dispatcher = _PHMMERDispatcher(
        queries=queries,
        targets=targets,
        cpus=cpus,
        backend=backend,
        callback=callback,  # type: ignore
        builder=builder,
        **options,
    )
    return dispatcher.run()  # type: ignore




import copy
import io
import itertools
import os
import shutil
import unittest
import tempfile

import pyhmmer
from pyhmmer.errors import EaselError, AlphabetMismatch
from pyhmmer.easel import Alphabet, SequenceFile, Randomness
from pyhmmer.plan7 import HMM, HMMFile, Profile, Background


class TestProfile(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        rng = Randomness(seed=0)
        cls.alphabet = Alphabet.amino()
        cls.hmm = HMM.sample(cls.alphabet, 100, rng)
        cls.hmm.name = b"hmm_one"
        cls.hmm.accession = b"HMM1"
        cls.hmm.description = b"first HMM"
        cls.background = Background(cls.alphabet)
        cls.profile = cls.hmm.to_profile(background=cls.background, L=200)

    def test_configure(self):
        profile = Profile(self.hmm.M, Alphabet.dna())
        bg = Background(profile.alphabet)
        self.assertNotEqual(profile.alphabet, self.hmm.alphabet)
        self.assertRaises(AlphabetMismatch, profile.configure, self.hmm, bg, 200)
        hmm = HMM(profile.alphabet, 100, b"test")
        self.assertNotEqual(profile.alphabet, self.background.alphabet)
        self.assertRaises(AlphabetMismatch, profile.configure, hmm, self.background, 200)

    def test_eq(self):
        profile2 = self.hmm.to_profile(background=self.background, L=200)
        self.assertEqual(self.profile, profile2)
        self.assertNotEqual(self.profile, 1)

    def test_copy(self):
        profile2 = copy.copy(self.profile)
        self.assertEqual(self.profile, profile2)

    def test_profile_modes(self):
        profile = self.profile.copy()
        for multihit in (False, True):
            for local in (False, True):
                profile.configure(
                    self.hmm,
                    self.background,
                    multihit=multihit,
                    local=local
                )
                self.assertEqual(profile.multihit, multihit)
                self.assertEqual(profile.local, local)

    def test_M(self):
        self.assertEqual(self.profile.M, self.hmm.M)

    def test_L(self):
        profile = self.hmm.to_profile(background=self.background, L=200)
        self.assertEqual(profile.L, 200)
        profile.L = 300
        self.assertEqual(profile.L, 300)

    def test_accession(self):
        self.assertEqual(self.profile.accession, self.hmm.accession)

    def test_description(self):
        self.assertEqual(self.profile.description, self.hmm.description)

    def test_name(self):
        self.assertEqual(self.profile.name, self.hmm.name)

    def test_clear(self):
        profile = self.profile.copy()
        self.assertNotEqual(profile.M, 0)
        profile.clear()
        self.assertEqual(profile.M, 0)

    def test_consensus(self):
        self.assertEqual(self.profile.consensus, self.hmm.consensus)
        profile = Profile(self.hmm.M, self.alphabet)
        self.assertIs(profile.consensus, None)

    def test_consensus_structure(self):
        self.assertEqual(self.profile.consensus, self.hmm.consensus)
        profile = Profile(self.hmm.M, self.alphabet)
        self.assertIs(profile.consensus, None)

    def test_offsets(self):
        self.assertIs(self.profile.offsets.model, None)
        self.assertIs(self.profile.offsets.profile, None)
        self.assertIs(self.profile.offsets.filter, None)



from libhmmer.p7_hmm cimport P7_HMM
from libhmmer.p7_bg cimport P7_BG


cdef extern from "hmmer.h" nogil:

    double p7_MeanMatchInfo(const P7_HMM* hmm, const P7_BG* bg);
    double p7_MeanMatchEntropy(const P7_HMM *hmm)
    double p7_MeanMatchRelativeEntropy(const P7_HMM *hmm, const P7_BG *bg)
    double p7_MeanForwardScore(const P7_HMM *hmm, const P7_BG *bg)
    int p7_MeanPositionRelativeEntropy(const P7_HMM *hmm, const P7_BG *bg, double *ret_entropy)
    int p7_hmm_CompositionKLD(const P7_HMM *hmm, const P7_BG *bg, float *ret_KL, float **opt_avp)



import unittest

from pyhmmer.easel import Alphabet
from pyhmmer.errors import (
    UnexpectedError,
    AllocationError,
    EaselError,
    AlphabetMismatch,
    InvalidParameter,
)


class TestErrors(unittest.TestCase):

    def test_unexpected_error(self):
        err = UnexpectedError(1, "p7_ReconfigLength")
        self.assertEqual(repr(err), "UnexpectedError(1, 'p7_ReconfigLength')")
        self.assertEqual(str(err), "Unexpected error occurred in 'p7_ReconfigLength': eslFAIL (status code 1)")

    def test_allocation_error(self):
        err = AllocationError("ESL_SQ", 16)
        self.assertEqual(repr(err), "AllocationError('ESL_SQ', 16)")
        self.assertEqual(str(err), "Could not allocate 16 bytes for type ESL_SQ")

        err2 = AllocationError("float", 4, 32)
        self.assertEqual(repr(err2), "AllocationError('float', 4, 32)")
        self.assertEqual(str(err2), "Could not allocate 128 bytes for an array of 32 float")

    def test_easel_error(self):
        err = EaselError(1, "failure")
        self.assertEqual(repr(err), "EaselError(1, 'failure')")
        self.assertEqual(str(err), "Error raised from C code: failure, eslFAIL (status code 1)")

    def test_alphabet_mismatch(self):
        err = AlphabetMismatch(Alphabet.dna(), Alphabet.rna())
        self.assertEqual(repr(err), "AlphabetMismatch(Alphabet.dna(), Alphabet.rna())")
        self.assertEqual(str(err), "Expected DNA alphabet, found RNA alphabet")
        self.assertNotEqual(err, 1)

        err2 = AlphabetMismatch(Alphabet.dna(), Alphabet.rna())
        self.assertEqual(err, err)
        self.assertEqual(err, err2)

        err3 = AlphabetMismatch(Alphabet.dna(), Alphabet.amino())
        self.assertNotEqual(err, err3)

    def test_invalid_parameters(self):
        err = InvalidParameter("x", 1)
        self.assertEqual(str(err), "Invalid 'x' parameter value: 1")

        err = InvalidParameter("x", 1, hint="positive number")
        self.assertEqual(str(err), "Invalid 'x' parameter value: 1 (expected positive number)")

        err = InvalidParameter("x", 1, choices=["x", "y", "z", None])
        self.assertEqual(str(err), "Invalid 'x' parameter value: 1 (expected 'x', 'y', 'z' or None)")


import pickle
import unittest
import sys

from pyhmmer.easel import (
    Matrix, 
    MatrixD, 
    MatrixF, 
    MatrixU8, 
    Vector, 
    VectorD, 
    VectorF, 
    VectorU8
)


class _TestMatrixBase(object):

    Matrix = NotImplemented

    @unittest.skipIf(sys.implementation.name == "pypy", "`getsizeof` not supported on PyPY")
    def test_sizeof(self):
        m1 = self.Matrix([[1, 2], [3, 4]])
        m2 = self.Matrix([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        self.assertGreater(sys.getsizeof(m2), sys.getsizeof(m1))

    def test_pickle(self):
        m1 = self.Matrix([[1, 2], [3, 4]])
        m2 = pickle.loads(pickle.dumps(m1))
        self.assertEqual(m1.shape, m2.shape)
        self.assertSequenceEqual(memoryview(m1), memoryview(m2))
        # self.assertSequenceEqual(m1, m2)
        for i in range(m1.shape[0]):
            self.assertEqual(m1[i], m2[i])
            for j in range(m1.shape[1]):
                self.assertEqual(m1[i,j], m2[i,j])

    def test_pickle_protocol4(self):
        m1 = self.Matrix([[1, 2], [3, 4]])
        m2 = pickle.loads(pickle.dumps(m1, protocol=4))
        self.assertEqual(m1.shape, m2.shape)
        self.assertSequenceEqual(memoryview(m1), memoryview(m2))
        # self.assertSequenceEqual(m1, m2)
        for i in range(m1.shape[0]):
            self.assertEqual(m1[i], m2[i])
            for j in range(m1.shape[1]):
                self.assertEqual(m1[i,j], m2[i,j])

    @unittest.skipUnless(sys.version_info >= (3, 8), "pickle protocol 5 requires Python 3.8+")
    def test_pickle_protocol5(self):
        m1 = self.Matrix([[1, 2], [3, 4]])
        m2 = pickle.loads(pickle.dumps(m1, protocol=5))
        self.assertEqual(m1.shape, m2.shape)
        self.assertSequenceEqual(memoryview(m1), memoryview(m2))
        # self.assertSequenceEqual(m1, m2)
        for i in range(m1.shape[0]):
            self.assertEqual(m1[i], m2[i])
            for j in range(m1.shape[1]):
                self.assertEqual(m1[i,j], m2[i,j])

    def test_empty_matrix(self):
        m1 = self.Matrix([])
        m2 = self.Matrix.zeros(0, 0)
        self.assertEqual(m1, m2)
        self.assertFalse(m2)
        self.assertFalse(m1)

    def test_init(self):
        mat = self.Matrix([ [1, 2], [3, 4] ])
        self.assertEqual(mat[0, 0], 1)
        self.assertEqual(mat[0, 1], 2)
        self.assertEqual(mat[1, 0], 3)
        self.assertEqual(mat[1, 1], 4)

    def test_init_error(self):
        self.assertRaises(ValueError, self.Matrix, [ [1.0, 2.0], [1.0] ])
        self.assertRaises(TypeError, self.Matrix, 1)
        self.assertRaises(TypeError, self.Matrix.zeros, [])

    def test_zeros(self):
        m = self.Matrix.zeros(2, 2)
        self.assertEqual(m.shape, (2, 2))
        self.assertEqual(m, self.Matrix([ [0, 0], [0, 0] ]))

        m2 = self.Matrix.zeros(4, 3)
        self.assertEqual(m2.shape, (4, 3))

        m3 = self.Matrix.zeros(0, 0)
        self.assertEqual(m3.shape, (0, 0))

        m4 = self.Matrix.zeros(0, 2)
        self.assertEqual(m4.shape, (0, 2))

        self.assertRaises(ValueError, self.Matrix.zeros, -2, 2)
        self.assertRaises(ValueError, self.Matrix.zeros, 2, -2)
        self.assertRaises(ValueError, self.Matrix.zeros, 0, -2)

    def test_eq(self):
        m1 = self.Matrix([ [1,2], [3,4] ])
        m2 = self.Matrix([ [1,2], [3,4] ])
        self.assertEqual(m1, m2)
        m3 = self.Matrix([ [2,2], [3,4] ])
        self.assertNotEqual(m1, m3)
        m4 = self.Matrix([ [1,2], [3,4], [5,6] ])
        self.assertNotEqual(m1, m4)

    def test_len(self):
        mat = self.Matrix([[1, 2], [3, 4]])
        self.assertEqual(len(mat), 2)
        mat = self.Matrix.zeros(10, 10)
        self.assertEqual(len(mat), 10)

    def test_min(self):
        mat = self.Matrix([ [1, 2], [3, 4] ])
        self.assertEqual(mat.min(), 1)

    def test_max(self):
        mat = self.Matrix([ [1, 2], [3, 4] ])
        self.assertEqual(mat.max(), 4)

    def test_argmin(self):
        mat = self.Matrix([ [1, 2], [3, 4] ])
        self.assertEqual(mat.argmin(), (0, 0))
        mat = self.Matrix([ [2, 1], [4, 3] ])
        self.assertEqual(mat.argmin(), (0, 1))

    def test_argmax(self):
        mat = self.Matrix([ [1, 2], [3, 4] ])
        self.assertEqual(mat.argmax(), (1, 1))
        mat = self.Matrix([ [2, 1], [4, 3] ])
        self.assertEqual(mat.argmax(), (1, 0))

    def test_copy(self):
        mat = self.Matrix([ [1, 2], [3, 4] ])
        mat2 = mat.copy()
        del mat
        self.assertIsInstance(mat2, self.Matrix)
        self.assertEqual(mat2[0, 0], 1)
        self.assertEqual(mat2[0, 1], 2)
        self.assertEqual(mat2[1, 0], 3)
        self.assertEqual(mat2[1, 1], 4)

    def test_add_matrix(self):
        m1 = self.Matrix([ [1, 2], [3, 4] ])
        m2 = self.Matrix([ [2, 2], [3, 3] ])
        m3 = m1 + m2
        self.assertEqual(m3[0, 0], 3)
        self.assertEqual(m3[0, 1], 4)
        self.assertEqual(m3[1, 0], 6)
        self.assertEqual(m3[1, 1], 7)

    def test_add_scalar(self):
        m1 = self.Matrix([ [2, 2], [3, 3] ])
        m2 = m1 + 2.0
        self.assertEqual(m2[0, 0], 4)
        self.assertEqual(m2[0, 1], 4)
        self.assertEqual(m2[1, 0], 5)
        self.assertEqual(m2[1, 1], 5)

    def test_iadd_scalar(self):
        mat = self.Matrix([ [1, 2], [3, 4] ])
        mat += 3
        self.assertEqual(mat[0, 0], 4)
        self.assertEqual(mat[0, 1], 5)
        self.assertEqual(mat[1, 0], 6)
        self.assertEqual(mat[1, 1], 7)

    def test_iadd_matrix(self):
        mat = self.Matrix([ [4, 5], [6, 7] ])
        mat += self.Matrix([ [2, 2], [3, 3] ])
        self.assertEqual(mat[0, 0], 6)
        self.assertEqual(mat[0, 1], 7)
        self.assertEqual(mat[1, 0], 9)
        self.assertEqual(mat[1, 1], 10)

    def test_mul_scalar(self):
        mat = self.Matrix([ [1, 2], [3, 4] ])
        m2 = mat * 3
        self.assertEqual(m2[0, 0], 3)
        self.assertEqual(m2[0, 1], 6)
        self.assertEqual(m2[1, 0], 9)
        self.assertEqual(m2[1, 1], 12)

    def test_mul_matrix(self):
        mat = self.Matrix([ [3, 6], [9, 12] ])
        m2 = mat * self.Matrix([ [2, 2], [3, 3] ])
        self.assertEqual(m2[0, 0], 6)
        self.assertEqual(m2[0, 1], 12)
        self.assertEqual(m2[1, 0], 27)
        self.assertEqual(m2[1, 1], 36)

    def test_imul_scalar(self):
        mat = self.Matrix([ [1, 2], [3, 4] ])
        mat *= 3
        self.assertEqual(mat[0, 0], 3)
        self.assertEqual(mat[0, 1], 6)
        self.assertEqual(mat[1, 0], 9)
        self.assertEqual(mat[1, 1], 12)

    def test_imul_matrix(self):
        mat = self.Matrix([ [3, 6], [9, 12] ])
        mat *= self.Matrix([ [2, 2], [3, 3] ])
        self.assertEqual(mat[0, 0], 6)
        self.assertEqual(mat[0, 1], 12)
        self.assertEqual(mat[1, 0], 27)
        self.assertEqual(mat[1, 1], 36)

    def test_sum(self):
        mat = self.Matrix([ [1, 2], [3, 4] ])
        self.assertEqual(mat.sum(), 1 + 2 + 3 + 4)

    def test_getitem_slice(self):
        mat = self.Matrix([ [1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16] ])

        m2 = mat[:2]
        self.assertEqual(m2, self.Matrix([[1, 2, 3, 4], [5, 6, 7, 8]]))

        m3 = mat[1:3]
        self.assertEqual(m3, self.Matrix([[5, 6, 7, 8], [9, 10, 11, 12]]))

        m4 = mat[-3:-1]
        self.assertEqual(m4, self.Matrix([[5, 6, 7, 8], [9, 10, 11, 12]]))

        m5 = mat[1:]
        self.assertEqual(m5, self.Matrix([[5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]]))

        m6 = mat[:]
        self.assertEqual(m6, mat)


class TestMatrix(unittest.TestCase):

    def test_abstract(self):
        self.assertRaises(TypeError, Matrix, [ [1, 2], [3, 4] ])
        self.assertRaises(TypeError, Matrix.zeros, 2, 2)


class TestMatrixF(_TestMatrixBase, unittest.TestCase):

    Matrix = MatrixF

    def test_memoryview_tolist(self):
        mat = MatrixF([ [1, 2], [3, 4] ])
        mem = memoryview(mat)
        self.assertEqual(mem.tolist(), [ [1.0, 2.0], [3.0, 4.0] ])

    def test_getitem_row(self):
        mat = MatrixF([ [1, 2], [3, 4] ])

        self.assertIsInstance(mat[0], VectorF)
        self.assertEqual(list(mat[0]), [1.0, 2.0])
        self.assertEqual(list(mat[-1]), [3.0, 4.0])

        with self.assertRaises(IndexError):
            row = mat[-10]

    def test_getitem_element(self):
        mat = MatrixF([ [1, 2], [3, 4] ])

        self.assertEqual(mat[0, 0], 1.0)
        self.assertEqual(mat[0, 1], 2.0)
        self.assertEqual(mat[1, 0], 3.0)
        self.assertEqual(mat[1, 1], 4.0)

        self.assertEqual(mat[0, -1], 2.0)
        self.assertEqual(mat[-1, 0], 3.0)
        self.assertEqual(mat[-1, -1], 4.0)

        with self.assertRaises(IndexError):
            x = mat[-10, 0]
        with self.assertRaises(IndexError):
            x = mat[0, -10]
        with self.assertRaises(IndexError):
            x = mat[0, 10]
        with self.assertRaises(IndexError):
            x = mat[10, 0]

    def test_from_raw_bytes_littleendian(self):
        mat = self.Matrix._from_raw_bytes(b'\x00\x00\x80>', 1, 1, byteorder="little")
        self.assertEqual(len(mat), 1)
        self.assertEqual(mat[0][0], 0.25)

    def test_from_raw_bytes_bigendian(self):
        mat = self.Matrix._from_raw_bytes(b'>\x80\x00\x00', 1, 1, byteorder="big")
        self.assertEqual(len(mat), 1)
        self.assertEqual(mat[0][0], 0.25)


class TestMatrixD(_TestMatrixBase, unittest.TestCase):

    Matrix = MatrixD

    def test_memoryview_tolist(self):
        mat = MatrixD([ [1, 2], [3, 4] ])
        mem = memoryview(mat)
        self.assertEqual(mem.tolist(), [ [1.0, 2.0], [3.0, 4.0] ])

    def test_getitem_row(self):
        mat = MatrixD([ [1, 2], [3, 4] ])

        self.assertIsInstance(mat[0], VectorD)
        self.assertEqual(list(mat[0]), [1.0, 2.0])
        self.assertEqual(list(mat[-1]), [3.0, 4.0])

        with self.assertRaises(IndexError):
            row = mat[-10]

    def test_getitem_element(self):
        mat = MatrixD([ [1, 2], [3, 4] ])

        self.assertEqual(mat[0, 0], 1.0)
        self.assertEqual(mat[0, 1], 2.0)
        self.assertEqual(mat[1, 0], 3.0)
        self.assertEqual(mat[1, 1], 4.0)

        self.assertEqual(mat[0, -1], 2.0)
        self.assertEqual(mat[-1, 0], 3.0)
        self.assertEqual(mat[-1, -1], 4.0)

        with self.assertRaises(IndexError):
            x = mat[-10, 0]
        with self.assertRaises(IndexError):
            x = mat[0, -10]
        with self.assertRaises(IndexError):
            x = mat[0, 10]
        with self.assertRaises(IndexError):
            x = mat[10, 0]

    def test_from_raw_bytes_littleendian(self):
        mat = self.Matrix._from_raw_bytes(b'\x00\x00\x00\x00\x00\x00\xd0\x3f', 1, 1, byteorder="little")
        self.assertEqual(len(mat), 1)
        self.assertEqual(mat[0][0], 0.25)

    def test_from_raw_bytes_bigendian(self):
        mat = self.Matrix._from_raw_bytes(b'\x3f\xd0\x00\x00\x00\x00\x00\x00', 1, 1, byteorder="big")
        self.assertEqual(len(mat), 1)
        self.assertEqual(mat[0][0], 0.25)


class TestMatrixU8(_TestMatrixBase, unittest.TestCase):

    Matrix = MatrixU8

    def test_memoryview_tolist(self):
        mat = MatrixU8([ [1, 2], [3, 4] ])
        mem = memoryview(mat)
        self.assertEqual(mem.tolist(), [ [1, 2], [3, 4] ])

    def test_getitem_row(self):
        mat = MatrixU8([ [1, 2], [3, 4] ])

        self.assertIsInstance(mat[0], VectorU8)
        self.assertEqual(list(mat[0]), [1, 2])
        self.assertEqual(list(mat[-1]), [3, 4])

        with self.assertRaises(IndexError):
            row = mat[-10]

    def test_getitem_element(self):
        mat = MatrixU8([ [1, 2], [3, 4] ])

        self.assertEqual(mat[0, 0], 1)
        self.assertEqual(mat[0, 1], 2)
        self.assertEqual(mat[1, 0], 3)
        self.assertEqual(mat[1, 1], 4)

        self.assertEqual(mat[0, -1], 2)
        self.assertEqual(mat[-1, 0], 3)
        self.assertEqual(mat[-1, -1], 4)

        with self.assertRaises(IndexError):
            x = mat[-10, 0]
        with self.assertRaises(IndexError):
            x = mat[0, -10]
        with self.assertRaises(IndexError):
            x = mat[0, 10]
        with self.assertRaises(IndexError):
            x = mat[10, 0]



# coding: utf-8
import typing

from pyhmmer.easel import Alphabet
from pyhmmer.plan7 import HMM

statuscode: typing.Dict[int, str]

class UnexpectedError(RuntimeError):
    code: int
    function: str
    def __init__(self, code: int, function: str) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...

class AllocationError(MemoryError):
    ctype: str
    itemsize: int
    count: int
    def __init__(self, ctype: str, itemsize: int, count: int = 1) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...

class EaselError(RuntimeError):
    code: int
    message: str
    def __init__(self, code: int, message: str) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...

class AlphabetMismatch(ValueError):
    expected: Alphabet
    actual: typing.Optional[Alphabet]
    def __init__(self, expected: Alphabet, actual: typing.Optional[Alphabet]) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __eq__(self, other: object) -> bool: ...

class ServerError(RuntimeError):
    code: int
    message: str
    def __init__(self, code: int, message: str) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...


class MissingCutoffs(ValueError):
    def __init__(
        self,
        model_name: typing.Optional[str] = None,
        bit_cutoffs: typing.Optional[str] = None,
    ) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...


class InvalidParameter(ValueError):
    name: str
    value: object
    choices: typing.Optional[typing.List[str]]
    hint: typing.Optional[str]
    def __init__(
        self,
        name: str,
        value: object,
        *,
        choices: typing.Optional[typing.List[str]],
        hint: typing.Optional[str]
    ) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...


class InvalidHMM(ValueError):
    hmm: HMM
    message: str
    def __init__(self, hmm: HMM, message: str) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...


